import json

import pytest
from pydantic import ValidationError

from npe2 import PluginManifest
from npe2.manifest import _validators

# the docstrings here are used to assert the validation error that is printed.

SAMPLE_PLUGIN_NAME = "my-plugin"


def _mutator_invalid_package_name(data):
    """'invalid??' is not a valid python package name."""
    data["name"] = "invalid??"


def _mutator_invalid_package_name2(data):
    """'_invalid' is not a valid python package name."""
    data["name"] = "_invalid"


def _mutator_command_not_begin_with_package_name(data):
    """Commands identifiers must start with the current package name"""
    assert "contributions" in data
    c = data["contributions"]["commands"][0]["id"]
    data["contributions"]["commands"][0]["id"] = ".".join(
        ["not_packagename", *c.split(".")[1:]]
    )


def _mutator_python_name_no_colon(data):
    """'this.has.no.colon' is not a valid python_name."""
    assert "contributions" in data
    data["contributions"]["commands"][0]["python_name"] = "this.has.no.colon"


def _mutator_python_name_locals(data):
    """functions defined in local scopes are not yet supported"""
    assert "contributions" in data
    data["contributions"]["commands"][0]["python_name"] = "mod:func.<locals>.another"


def _mutator_python_name_starts_with_number(data):
    """'1starts_with_number' is not a valid python_name."""
    assert "contributions" in data
    data["contributions"]["commands"][0]["python_name"] = "1starts_with_number"


def _mutator_writer_requires_non_empty_layer_types(data):
    """layer_types must not be empty"""
    data["contributions"]["writers"][0]["layer_types"] = []


def _mutator_writer_invalid_layer_type_constraint(data):
    """'image{' is not a valid LayerType"""
    data["contributions"]["writers"][0]["layer_types"].append("image{")


def _mutator_writer_invalid_file_extension_1(data):
    """Invalid file extension: Must have one character past the '.'"""
    data["contributions"]["writers"][0]["filename_extensions"] = ["*"]


def _mutator_writer_invalid_file_extension_2(data):
    """Invalid file extension: Must have one character past the '.'"""
    data["contributions"]["writers"][0]["filename_extensions"] = ["."]


def _mutator_invalid_icon(data):
    """is not a valid icon URL. It must start with 'https://'"""
    data["icon"] = "http://example.com/icon.png"


@pytest.mark.parametrize(
    "mutator",
    [
        _mutator_invalid_package_name,
        _mutator_invalid_package_name2,
        _mutator_command_not_begin_with_package_name,
        _mutator_python_name_no_colon,
        _mutator_python_name_locals,
        _mutator_python_name_starts_with_number,
        _mutator_writer_requires_non_empty_layer_types,
        _mutator_writer_invalid_layer_type_constraint,
        _mutator_writer_invalid_file_extension_1,
        _mutator_writer_invalid_file_extension_2,
        _mutator_invalid_icon,
    ],
)
def test_invalid(mutator, uses_sample_plugin):
    result = next(
        result
        for result in PluginManifest.discover()
        if result.manifest and result.manifest.name == SAMPLE_PLUGIN_NAME
    )
    assert result.error is None
    assert result.manifest is not None
    pm = result.manifest
    data = json.loads(pm.json(exclude_unset=True))
    mutator(data)
    with pytest.raises(ValidationError) as excinfo:
        PluginManifest(**data)
    assert mutator.__doc__ in str(excinfo.value)


def test_schema_version_too_high():
    with pytest.warns(
        UserWarning, match=r"\(999.999.999\) is newer than npe2's schema version"
    ):
        PluginManifest(name="sample", schema_version="999.999.999")


def test_invalid_python_name(uses_sample_plugin):
    mf = next(
        result
        for result in PluginManifest.discover()
        if result.manifest and result.manifest.name == SAMPLE_PLUGIN_NAME
    ).manifest
    assert mf and mf.contributions and mf.contributions.commands
    assert mf.contributions.commands[-1].python_name

    mf.validate_imports()
    mf.contributions.commands[-1].python_name += "_whoops"  # type: ignore
    with pytest.raises(ValidationError) as e:
        mf.validate_imports()
    assert "has no attribute 'make_widget_from_function_whoops'" in str(e.value)


def _valid_mutator_no_contributions(data):
    """
    Contributions can be absent, in which case the Pydantic model will set the
    default value to an empty Contributions model
    """
    del data["contributions"]


def _valid_mutator_no_contributions_empty(data):
    """
    Contributions can be an empty list, in which case the Pydantic model will set the
    default value to an empty Contributions model
    """
    data["contributions"] = []


def _valid_mutator_no_contributions_None(data):
    """
    Contributions can be None, in which case the Pydantic model will set the
    default value to None, and not the empty list, make sure that works.
    """
    # This is no longer recommended.  A missing contributions is fine, and an empty
    # list is fine.  We preserve this for backwards compatibility,
    # but providing None explicitly shouldn't be used
    data["contributions"] = None


@pytest.mark.parametrize(
    "mutator",
    [_valid_mutator_no_contributions, _valid_mutator_no_contributions_None],
)
def test_valid_mutations(mutator, uses_sample_plugin):
    assert mutator.__name__.startswith("_valid")

    pm = list(PluginManifest.discover())[0]
    assert pm.manifest
    # make sure the data is a copy as we'll mutate it
    data = json.loads(pm.manifest.json(exclude_unset=True))
    mutator(data)
    PluginManifest(**data)


@pytest.mark.parametrize(
    "display_name",
    [
        "Here there everywhere and more with giggles and friends on top of a mountainside drinking tea",  # noqa
        "ab",
        "a   ",
        " abc",
        "_abc",
        "abc♱",
    ],
)
def test_invalid_display_names(display_name, uses_sample_plugin):
    field = PluginManifest.__fields__["display_name"]
    value, err = field.validate(display_name, {}, loc="display_name")
    assert err is not None


@pytest.mark.parametrize(
    "display_name",
    [
        "Some Cell & Stru买cture Segmenter",
        "Segment Blobs and Things with Membranes",
        "Segment: and Things.2 with Membranes ~= 8",
        "abc",
        "abc䜁䜂",
    ],
)
def test_valid_display_names(display_name, uses_sample_plugin):
    field = PluginManifest.__fields__["display_name"]
    value, err = field.validate(display_name, {}, loc="display_name")
    assert err is None


def test_display_name_default_is_valid():
    PluginManifest(name="")


@pytest.mark.parametrize(
    "expr",
    [
        "vectors{",
        "image",  # should parse fine, but be a duplication error
        "vectors{8,3}",
        "vectors{-1}",
        "vectors??",
        "other?",
    ],
)
def test_writer_invalid_layer_type_expressions(expr, uses_sample_plugin):
    result = next(
        result
        for result in PluginManifest.discover()
        if result.manifest and result.manifest.name == SAMPLE_PLUGIN_NAME
    )
    assert result.error is None
    assert result.manifest is not None
    pm = result.manifest
    data = json.loads(pm.json(exclude_unset=True))

    assert "contributions" in data
    assert "writers" in data["contributions"]
    data["contributions"]["writers"][0]["layer_types"].append(expr)

    with pytest.raises(ValidationError):
        PluginManifest(**data)


@pytest.mark.parametrize("id", ["badchar!?", "-bad-start", "has space"])
def test_invalid_command_id(id):
    with pytest.raises(ValueError):
        _validators.command_id(id)



# mypy: disable-error-code=empty-body
"""This module mimics all of the contributions my-plugin...
but is used to reverse-engineer the manifest."""
from typing import TYPE_CHECKING, Any, Dict, List, Tuple

# to test various ways that this can be imported, since we're using static parsing.
import npe2.implements
import npe2.implements as impls
from npe2 import implements
from npe2.implements import reader

# alternative pattern that does not require npe2 at runtime
if TYPE_CHECKING:
    from npe2 import implements as noimport
else:
    # create no-op `implements.anything(**kwargs)` decorator
    D = type("D", (), {"__getattr__": lambda *_: (lambda **_: (lambda f: f))})
    noimport = D()


@implements.on_activate
def activate(ctx):
    ...


@implements.on_deactivate
def deactivate(ctx):
    ...


@implements.reader(
    id="some_reader",
    title="Some Reader",
    filename_patterns=["*.fzy", "*.fzzy"],
    accepts_directories=True,
)
def get_reader(path: str):
    ...


@reader(
    id="url_reader",
    title="URL Reader",
    filename_patterns=["http://*", "https://*"],
    accepts_directories=False,
    ensure_args_valid=True,
)
def url_reader(path: str):
    ...


@noimport.writer(
    id="my_writer",
    title="My Multi-layer Writer",
    filename_extensions=["*.tif", "*.tiff"],
    layer_types=["image{2,4}", "tracks?"],
)
@implements.writer(
    id="my_writer",
    title="My Multi-layer Writer",
    filename_extensions=["*.pcd", "*.e57"],
    layer_types=["points{1}", "surface+"],
)
def writer_function(path: str, layer_data: List[Tuple[Any, Dict, str]]) -> List[str]:
    ...


@implements.writer(
    id="my_single_writer",
    title="My single-layer Writer",
    filename_extensions=["*.xyz"],
    layer_types=["labels"],
)
def writer_function_single(path: str, layer_data: Any, meta: Dict) -> List[str]:
    ...


@npe2.implements.widget(
    id="some_widget", title="Create my widget", display_name="My Widget"
)
class SomeWidget:
    ...


@npe2.implements.sample_data_generator(
    id="my-plugin.generate_random_data",  # the plugin-name is optional
    title="Generate uniform random data",
    key="random_data",
    display_name="Some Random Data (512 x 512)",
)
def random_data():
    ...


@impls.widget(
    id="some_function_widget",
    title="Create widget from my function",
    display_name="A Widget From a Function",
    autogenerate=True,
)
def make_widget_from_function(x: int, threshold: int):
    ...



from typing import Optional, Union

from pydantic import BaseModel, Field

from npe2.manifest.utils import Executable


# user provides this
class _MenuItem(BaseModel):
    """Generic menu item contribution."""

    when: Optional[str] = Field(
        description="Condition which must be true to *show* this item in the menu. "
        "Note that ``when`` clauses apply to menus and ``enablement`` clauses to "
        "commands. The ``enablement`` applies to all menus and even keybindings while "
        "the ``when`` only applies to a single menu."
    )
    # TODO: declare groups for every menu exposed by napari:
    # e.g. `2_compare`, `4_search`, `6_cutcopypaste`
    group: Optional[str] = Field(
        description="The `group` property defines sorting and grouping of menu items. "
        "The `'navigation'` group is special: it will always be sorted to the "
        "top/beginning of a menu. By default, the order *inside* a group depends on "
        "the `title`. The group-local order of a menu item can be specified by "
        "appending @<int> to the group identifier: e.g. `group: 'myGroup@2'`."
    )


class Submenu(_MenuItem):
    """Contributes a submenu placement in a menu."""

    submenu: str = Field(
        ...,
        description="Identifier of the submenu to display in this item."
        "The submenu must be declared in the 'submenus' -section",
    )
    # if submenu doesn't exist, you get:
    # Menu item references a submenu ...` which is not defined in the 'submenus' section


class MenuCommand(_MenuItem, Executable):
    """Contributes a command in a menu."""

    command: str = Field(
        ...,
        description="Identifier of the command to execute. "
        "The command must be declared in the 'commands' section",
    )
    # if command doesn't exist, you get:
    # "Menu item references a command `...` which is not defined in the
    # 'commands' section."
    alt: Optional[str] = Field(
        description="Identifier of an alternative command to execute. "
        "It will be shown and invoked when pressing Alt while opening a menu."
        "The command must be declared in the 'commands' section"
    )


MenuItem = Union[MenuCommand, Submenu]



import builtins
import sys
import warnings
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, List, Optional, Sequence

import typer

from npe2 import PluginManager, PluginManifest, __version__

if TYPE_CHECKING:
    from rich.console import RenderableType

app = typer.Typer(no_args_is_help=True)


def _show_version_and_exit(value: bool) -> None:
    if value:
        typer.echo(f"npe2 v{__version__}")
        raise typer.Exit()


@app.callback()
def _main(
    version: Optional[bool] = typer.Option(
        None,
        "-v",
        "--version",
        callback=_show_version_and_exit,
        help="Show version and exit.",
        is_eager=True,
    ),
):
    """npe2: napari plugin engine (v{version})

    For additional help on a specific command: type 'npe2 [command] --help'
    """


_main.__doc__ = typer.style(
    (_main.__doc__ or "").format(version=__version__), fg="bright_yellow"
)

SYNTAX_THME = "monokai"
SYNTAX_BACKGROUND = "black"


class Format(str, Enum):
    """Valid manifest file formats."""

    yaml = "yaml"
    json = "json"
    toml = "toml"


class ListFormat(str, Enum):
    """Valid out formats for `npe2 list`."""

    table = "table"
    json = "json"  # alias for json in pandas "records" format
    yaml = "yaml"
    compact = "compact"


def _pprint_formatted(string, format: Format = Format.yaml):  # pragma: no cover
    """Print yaml nicely, depending on available modules."""
    from rich.console import Console
    from rich.syntax import Syntax

    syntax = Syntax(
        string, format.value, theme=SYNTAX_THME, background_color=SYNTAX_BACKGROUND
    )
    Console().print(syntax)


def _pprint_exception(err: Exception):
    from rich.console import Console
    from rich.traceback import Traceback

    e_info = (type(err), err, err.__traceback__)

    trace = Traceback.extract(*e_info, show_locals=True)
    Console().print(Traceback(trace))


def _pprint_table(
    headers: Sequence["RenderableType"], rows: Sequence[Sequence["RenderableType"]]
):
    from itertools import cycle

    from rich.console import Console
    from rich.table import Table

    COLORS = ["cyan", "magenta", "green", "yellow"]
    EMOJI_TRUE = ":white_check_mark:"
    EMOJI_FALSE = ""

    table = Table()
    for head, color in zip(headers, cycle(COLORS)):
        table.add_column(head, style=color)
    for row in rows:
        strings = []
        for r in row:
            val = ""
            if isinstance(r, dict):
                val = ", ".join(f"{k} ({v})" for k, v in r.items())
            elif r:
                val = str(r).replace("True", EMOJI_TRUE).replace("False", EMOJI_FALSE)
            strings.append(val)
        table.add_row(*strings)

    console = Console()
    console.print(table)


@app.command()
def validate(
    name: str,
    imports: bool = typer.Option(
        False,
        help="Validate all `python_name` entries by importing. This runs python code! "
        "package must be importable on sys.path.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        help="Print tracebacks on error.",
    ),
):
    """Validate manifest for a distribution name or manifest filepath."""
    err: Optional[Exception] = None
    try:
        pm = PluginManifest._from_package_or_name(name)
        msg = f"✔ Manifest for {(pm.display_name or pm.name)!r} valid!"
        if imports:
            pm.validate_imports()
    except PluginManifest.ValidationError as e:
        msg = f"🅇 Invalid! {e}"
        err = e
    except Exception as e:
        msg = f"🅇 Unexpected error in {name!r}.\n{type(e).__name__}: {e}"
        err = e

    typer.secho(msg, fg=typer.colors.RED if err else typer.colors.GREEN, bold=True)
    if err is not None:
        if debug:
            _pprint_exception(err)

        raise typer.Exit(1)


def _check_output(output: Path) -> Format:
    if output.suffix.lstrip(".") not in Format._member_names_:
        typer.echo(
            f"Invalid output extension {output.suffix!r}. Must be one of: "
            + ", ".join(Format._member_names_)
        )
        raise typer.Exit(1)
    return Format(output.suffix.lstrip("."))


@app.command()
def parse(
    name: str = typer.Argument(
        ..., help="Name of an installed package, or path to a manifest file."
    ),
    format: Format = typer.Option(
        "yaml", "-f", "--format", help="Markdown format to use."
    ),
    indent: Optional[int] = typer.Option(
        None,
        "--indent",
        help="Number of spaces to indent (for json)",
        min=0,
        max=10,
    ),
    output: Optional[Path] = typer.Option(
        None,
        "-o",
        "--output",
        exists=False,
        help="If provided, will write manifest to filepath (must end with .yaml, "
        ".json, or .toml). Otherwise, will print to stdout.",
    ),
):
    """Show parsed manifest as yaml."""
    fmt = _check_output(output) if output else format
    pm = PluginManifest._from_package_or_name(name)
    manifest_string = getattr(pm, fmt.value)(indent=indent)
    if output:
        output.write_text(manifest_string)
    else:
        _pprint_formatted(manifest_string, fmt)


def _make_rows(pm_dict: dict, normed_fields: Sequence[str]) -> Iterator[List]:
    """Cleanup output from pm.dict() into rows for table.

    outside of just extracting the fields we care about, this also:

    - handles nested fields expressed as dotted strings: `packge_metadata.version`
    - negates fields that are prefixed with `!`
    - simplifies contributions to a {name: count} dict.

    """
    for info in pm_dict["plugins"].values():
        row = []
        for field in normed_fields:
            val = info.get(field.lstrip("!"))

            # extact nested fields
            if not val and "." in field:
                parts = field.split(".")
                val = info
                while parts and hasattr(val, "__getitem__"):
                    val = val[parts.pop(0)]

            # negate fields starting with !
            if field.startswith("!"):
                val = not val

            # simplify contributions to just the number of contributions
            if field == "contributions":
                val = {k: len(v) for k, v in val.items() if v}

            row.append(val)
        yield row


@app.command()
def list(
    fields: str = typer.Option(
        "name,version,npe2,contributions",
        help="Comma seperated list of fields to include in the output."
        "Names may contain dots, indicating nested manifest fields "
        "(`contributions.readers`). Fields names prefixed with `!` will be "
        "negated in the output. Fields will appear in the table in the order in "
        "which they are provided.",
        metavar="FIELDS",
    ),
    sort: str = typer.Option(
        "0",
        "-s",
        "--sort",
        help="Field name or (int) index on which to sort.",
        metavar="KEY",
    ),
    format: ListFormat = typer.Option(
        "table",
        "-f",
        "--format",
        help="Out format to use. When using 'compact', `--fields` is ignored ",
    ),
):
    """List currently installed plugins."""

    if format == ListFormat.compact:
        fields = "name,version,contributions"

    requested_fields = [f.lower() for f in fields.split(",")]

    # check for sort values that will not work
    bad_sort_param_msg = (
        f"Invalid sort value {sort!r}. "
        f"Must be column index (<{len(requested_fields)}) or one of: "
        + ", ".join(requested_fields)
    )
    try:
        if (sort_index := int(sort)) >= len(requested_fields):
            raise typer.BadParameter(bad_sort_param_msg)
    except ValueError:
        try:
            sort_index = requested_fields.index(sort.lower())
        except ValueError as e:
            raise typer.BadParameter(bad_sort_param_msg) from e

    # some convenience aliases
    ALIASES = {
        "version": "package_metadata.version",
        "summary": "package_metadata.summary",
        "license": "package_metadata.license",
        "author": "package_metadata.author",
        "npe2": "!npe1_shim",
        "npe1": "npe1_shim",
    }
    normed_fields = [ALIASES.get(f, f) for f in requested_fields]

    pm = PluginManager.instance()
    pm.discover(include_npe1=True)
    pm.index_npe1_adapters()
    pm_dict = pm.dict(include={f.lstrip("!") for f in normed_fields})
    rows = sorted(_make_rows(pm_dict, normed_fields), key=lambda r: r[sort_index])

    if format == ListFormat.table:
        heads = [f.split(".")[-1].replace("_", " ").title() for f in requested_fields]
        _pprint_table(headers=heads, rows=rows)
        return

    # standard records format used for the other formats
    # [{column -> value}, ... , {column -> value}]
    data: List[dict] = [dict(zip(requested_fields, row)) for row in rows]

    if format == ListFormat.json:
        import json

        _pprint_formatted(json.dumps(data, indent=1), Format.json)
    elif format in (ListFormat.yaml):
        import yaml

        _pprint_formatted(yaml.safe_dump(data, sort_keys=False), Format.yaml)
    elif format in (ListFormat.compact):
        template = "  - {name}: {version} ({ncontrib} contributions)"
        for r in data:
            ncontrib = sum(r.get("contributions", {}).values())
            typer.echo(template.format(**r, ncontrib=ncontrib))


def _fetch_all_manifests(doit: bool):
    """Fetch all manifests and dump to "manifests" folder."""
    if not doit:
        return

    from npe2._inspection import _fetch

    dest = "manifests"
    if "-o" in sys.argv:
        dest = sys.argv[sys.argv.index("-o") + 1]
    elif "--output" in sys.argv:  # pragma: no cover
        dest = sys.argv[sys.argv.index("--output") + 1]

    _fetch.fetch_all_manifests(dest)
    raise typer.Exit(0)


@app.command()
def fetch(
    name: List[str],
    version: Optional[str] = None,
    include_package_meta: Optional[bool] = typer.Option(
        False,
        "-m",
        "--include-package-meta",
        help="Include package metadata in the manifest.",
    ),
    format: Format = typer.Option(
        "yaml", "-f", "--format", help="Markdown format to use."
    ),
    indent: Optional[int] = typer.Option(
        None,
        "--indent",
        help="Number of spaces to indent (for json)",
        min=0,
        max=10,
    ),
    output: Optional[Path] = typer.Option(
        None,
        "-o",
        "--output",
        exists=False,
        help="If provided, will write manifest to filepath (must end with .yaml, "
        ".json, or .toml). Otherwise, will print to stdout.",
    ),
    all: Optional[bool] = typer.Option(
        None,
        "--all",
        help="Fetch manifests for ALL known plugins (will be SLOW)",
        callback=_fetch_all_manifests,
        is_eager=True,
    ),
):
    """Fetch manifest from remote package.

    If an npe2 plugin is detected, the manifest is returned directly, otherwise
    it will be installed into a temporary directory, imported, and discovered.
    """
    from npe2 import fetch_manifest

    fmt = _check_output(output) if output else format
    kwargs: dict = {"indent": indent}
    if include_package_meta:
        kwargs["exclude"] = set()

    for n in name:
        mf = fetch_manifest(n, version=version)
        manifest_string = getattr(mf, fmt.value)(**kwargs)

        if output:
            output.write_text(manifest_string, encoding="utf-8")
        else:
            _pprint_formatted(manifest_string, fmt)


@app.command()
def convert(
    path: Path = typer.Argument(
        ...,
        help="Path of a local repository to convert (package must also be installed in"
        " current environment). Or, the name of an installed package/plugin. If a "
        "package is provided instead of a directory, the new manifest will simply be "
        "printed to stdout.",
    ),
    dry_run: Optional[bool] = typer.Option(
        False,
        "--dry-runs",
        "-n",
        help="Just print manifest to stdout. Do not modify anything",
    ),
):
    """Convert first generation napari plugin to new (manifest) format."""
    from ._inspection._from_npe1 import convert_repository, manifest_from_npe1

    try:
        with warnings.catch_warnings(record=True) as w:
            if path.is_dir():
                pm, mf_path = convert_repository(path, dry_run=dry_run)
            else:
                pm = manifest_from_npe1(str(path))
        if w:
            from textwrap import indent

            typer.secho("Some issues occured:", fg=typer.colors.RED, bold=False)
            for r in w:
                typer.secho(
                    indent(str(r.message), "  "),
                    fg=typer.colors.MAGENTA,
                    bold=False,
                )
            typer.echo()

    except Exception as e:
        msg = f"Conversion failed:\n{type(e).__name__}: {e}"
        typer.secho(msg, fg=typer.colors.RED, bold=True)
        raise typer.Exit(1) from e

    if dry_run or not path.is_dir():
        if path.is_dir():
            typer.secho(
                f"# Manifest would be written to {mf_path}",
                fg=typer.colors.BRIGHT_GREEN,
                bold=False,
            )
        _pprint_formatted(pm.yaml(), Format.yaml)
    else:
        msg = f"✔  Conversion complete! New manifest at {mf_path}."
        typer.secho(msg, fg=typer.colors.GREEN, bold=True)
        typer.secho(
            "If you have any napari_plugin_engine imports or hook_implementation "
            "decorators, you may remove them now.",
            fg=typer.colors.GREEN,
            bold=False,
        )


@app.command()
def cache(
    clear: Optional[bool] = typer.Option(
        False, "--clear", "-d", help="Clear the npe1 adapter manifest cache"
    ),
    names: List[str] = typer.Argument(
        None, help="Name(s) of distributions to list/delete"
    ),
    list_: Optional[bool] = typer.Option(
        False, "--list", "-l", help="List cached manifests"
    ),
    verbose: Optional[bool] = typer.Option(False, "--verbose", "-v", help="verbose"),
):
    """Cache utils"""
    from npe2.manifest._npe1_adapter import ADAPTER_CACHE, clear_cache

    if clear:
        if _cleared := clear_cache(names):
            nf = "\n".join(f" - {i.name}" for i in _cleared)
            typer.secho("Cleared these files from cache:")
            typer.secho(nf, fg=typer.colors.RED)
        else:
            msg = "Nothing to clear"
            if names:
                msg += f" for plugins: {','.join(names)}"
            typer.secho(msg, fg=typer.colors.RED)

        typer.Exit()
    if list_:
        files = builtins.list(ADAPTER_CACHE.glob("*.yaml"))
        if names:
            files = [f for f in files if any(f.name.startswith(n) for n in names)]

        if not files:
            if names:
                typer.secho(f"Nothing cached for plugins: {','.join(names)}")
            else:
                typer.secho("Nothing cached")
            typer.Exit()
        for fname in files:
            mf = PluginManifest.from_file(fname)
            if verbose:
                _pprint_formatted(mf.yaml(), Format.yaml)  # pragma: no cover
            else:
                typer.secho(f"{mf.name}: {mf.package_version}", fg=typer.colors.GREEN)


@app.command()
def compile(
    src_dir: str,
    output: Optional[Path] = typer.Option(
        None,
        "-o",
        "--output",
        exists=False,
        help="If provided, will write manifest to filepath (must end with .yaml, "
        ".json, or .toml). Otherwise, will print to stdout.",
    ),
    format: Format = typer.Option(
        "yaml", "-f", "--format", help="Markdown format to use."
    ),
):
    """Compile @npe2.implements contributions to generate a manifest."""
    from . import _inspection

    manifest = _inspection.compile(src_dir, dest=output)
    manifest_string = getattr(manifest, format.value)(indent=2)
    _pprint_formatted(manifest_string, format)


def main():
    app()



# mypy: disable-error-code=empty-body
from typing import TYPE_CHECKING, Any, Dict, List, Tuple

# alternative pattern that does not require npe2 at runtime
if TYPE_CHECKING:
    from npe2 import implements
else:
    # create no-op `implements.anything(**kwargs)` decorator
    D = type("D", (), {"__getattr__": lambda *_: (lambda **_: (lambda f: f))})
    implements = D()


@implements.on_activate
def activate(ctx):
    ...


@implements.on_deactivate
def deactivate(ctx):
    ...


@implements.reader(
    id="some_reader",
    title="Some Reader",
    filename_patterns=["*.fzy", "*.fzzy"],
    accepts_directories=True,
)
def get_reader(path: str):
    ...


@implements.writer(
    id="my_writer",
    title="My Multi-layer Writer",
    filename_extensions=["*.tif", "*.tiff"],
    layer_types=["image{2,4}", "tracks?"],
)
@implements.writer(
    id="my_writer",
    title="My Multi-layer Writer",
    filename_extensions=["*.pcd", "*.e57"],
    layer_types=["points{1}", "surface+"],
)
def writer_function(path: str, layer_data: List[Tuple[Any, Dict, str]]) -> List[str]:
    ...



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List, Optional, Union

from pydantic.fields import Field
from pydantic.generics import GenericModel

from npe2.manifest.utils import Executable
from npe2.types import LayerData

if TYPE_CHECKING:
    from npe2._command_registry import CommandRegistry


class _SampleDataContribution(GenericModel, ABC):
    """Contribute sample data for use in napari.

    Sample data can take the form of a **command** that returns layer data, or a simple
    path or **uri** to a local or remote resource (assuming there is a reader plugin
    capable of reading that path/URI).
    """

    key: str = Field(..., description="A unique key to identify this sample.")
    display_name: str = Field(
        ..., description="String to show in the UI when referring to this sample"
    )

    @abstractmethod
    def open(
        self, *args, _registry: Optional["CommandRegistry"] = None, **kwargs
    ) -> List[LayerData]:
        ...


class SampleDataGenerator(_SampleDataContribution, Executable[List[LayerData]]):
    """Contribute a callable command that creates data on demand."""

    command: str = Field(
        ..., description="Identifier of a command that returns layer data tuple."
    )

    def open(
        self, *args, _registry: Optional["CommandRegistry"] = None, **kwargs
    ) -> List[LayerData]:
        return self.exec(args, kwargs, _registry=_registry)

    class Config:
        title = "Sample Data Function"


class SampleDataURI(_SampleDataContribution):
    """Contribute a URI to static local or remote data. This can be data included in
    the plugin package, or a URL to remote data.  The URI must be readable by either
    napari's builtin reader, or by a plugin that is included/required."""

    uri: str = Field(
        ...,
        description="Path or URL to a data resource. "
        "This URI should be a valid input to `io_utils.read`",
    )
    reader_plugin: Optional[str] = Field(
        None,
        description="Name of plugin to use to open URI",
    )

    def open(self, *args, **kwargs) -> List[LayerData]:
        from npe2.io_utils import read

        return read([self.uri], plugin_name=self.reader_plugin, stack=False)

    class Config:
        title = "Sample Data URI"


SampleDataContribution = Union[SampleDataGenerator, SampleDataURI]



from enum import Enum
from typing import List, Tuple

from pydantic import BaseModel, Extra, Field, validator

from npe2.manifest.utils import Executable


class LayerType(str, Enum):
    image = "image"
    labels = "labels"
    points = "points"
    shapes = "shapes"
    surface = "surface"
    tracks = "tracks"
    vectors = "vectors"


class LayerTypeConstraint(BaseModel):
    """Layer type constraints.

    A writer plugin can declare that it will write 0 or more layers of a
    specific type.

    For example:

    ```
        image      Write exactly 1 image layer.
        image?     Write 0 or 1 image layes.
        image+     Write 1 or more image layers.
        image*     Write 0 or more image layers.
        image{k}   Write exactly k image layres.
        image{m,n} Write between m and n layers (inclusive range). Must have m<=n.
    ```

    When a type is not present in the list of constraints, that
    corresponds to a writer that is not compatible with that type.

    For example, a writer declaring:

    ```
        layer_types=["image+", "points*"]
    ```

    would not be selected when trying to write an `image` and a `vector`
    layer because the above only works for cases with 0 `vector` layers.

    Note that just because a writer declares compatibility with a layer
    type does not mean it actually writes that type.  In the example
    above, the writer might accept a set of layers containing `image`s and
    `point`s, but the write command might just ignore the `point` layers
    """

    layer_type: LayerType
    bounds: Tuple[int, int] = Field(
        ...,
        description="This writer consumes between bounds[0] and bounds[1] "
        "layers of `layer_type`",
    )

    @validator("bounds")
    def check_bounds(cls, v):
        mn, mx = v
        assert mn >= 0, "min must be >= 0"
        assert mx > mn, "max must be > min"
        return v

    @classmethod
    def zero(cls, layer_type: LayerType) -> "LayerTypeConstraint":
        return cls(layer_type=layer_type, bounds=(0, 1))

    def is_zero(self) -> bool:
        return self.bounds == (0, 1)

    def max(self) -> int:
        return max(0, self.bounds[1] - 1)

    @classmethod
    def from_str(cls, expr: str) -> "LayerTypeConstraint":
        """Parse layer-type constraint expressions.

        These have the form '<layer_type><range>' where <range> is one of:
        '?', '+', '*', '{k}', '{m,n}'.

        '?' means 0 or 1.
        '+' means 1 or more.
        '*' means 0 or more.
        '{k}' means exactly k.
        '{m,n}' means between m and n (inclusive).
        """
        # Writers won't accept more than this number of layers.
        MAX_LAYERS = 1 << 32

        def parse(expr):
            if expr.endswith("?"):
                return (0, 2), LayerType(expr[:-1])
            elif expr.endswith("+"):
                return (1, MAX_LAYERS), LayerType(expr[:-1])
            elif expr.endswith("*"):
                return (0, MAX_LAYERS), LayerType(expr[:-1])
            elif expr.endswith("}"):
                rest, _, range_expr = expr[:-1].rpartition("{")
                if "," in range_expr:
                    m, n = range_expr.split(",")
                    return (int(m), int(n) + 1), LayerType(rest)
                else:
                    k = int(range_expr)
                    return (k, k + 1), LayerType(rest)
            else:
                return (1, 2), LayerType(expr)

        bounds, lt = parse(expr)
        return cls(layer_type=lt, bounds=bounds)


class WriterContribution(Executable[List[str]]):
    r"""Contribute a layer writer.

    Writers accept data from one or more layers and write them to file. Writers declare
    support for writing one or more **layer_types**, may be associated with specific
    **filename_patterns** (e.g. "\*.tif", "\*.zip") and are invoked whenever
    `viewer.layers.save('some/path.ext')` is used on the command line, or when a user
    requests to save one or more layers in the graphical user interface with `File ->
    Save Selected Layer(s)...` or `Save All Layers...`
    """

    command: str = Field(
        ..., description="Identifier of the command providing a writer."
    )
    layer_types: List[str] = Field(
        ...,
        description="List of layer type constraints. These determine what "
        "layers (or combinations thereof) this writer handles.",
    )
    # An empty filename_extensions list matches any file extension. Making the
    # default something like ['.*'] is tempting but we don't actually use
    # these for glob matching and supporting this default ends up making the
    # code more complicated.
    filename_extensions: List[str] = Field(
        default_factory=list,
        description="List of filename extensions compatible with this writer. "
        "The first entry is used as the default if necessary. Empty by default. "
        "When empty, any filename extension is accepted.",
    )
    display_name: str = Field(
        default="",
        description="Brief text used to describe this writer when presented. "
        "Empty by default. When present, this string is presented in the save dialog "
        "along side the plugin name and may be used to distinguish the kind of "
        "writer for the user. E.g. “lossy” or “lossless”.",
    )

    def layer_type_constraints(self) -> List[LayerTypeConstraint]:
        spec = [LayerTypeConstraint.from_str(lt) for lt in self.layer_types]
        unspecified_types = set(LayerType) - {lt.layer_type for lt in spec}
        return spec + [LayerTypeConstraint.zero(lt) for lt in unspecified_types]

    def __hash__(self):
        return hash(
            (
                self.command,
                str(self.layer_types),
                str(self.filename_extensions),
                self.display_name,
            )
        )

    class Config:
        extra = Extra.forbid

    @validator("layer_types")
    def _parsable_layer_type_expr(cls, layer_types: List[str]) -> List[str]:
        try:
            # a successful parse means the string is valid
            for lt in layer_types:
                LayerTypeConstraint.from_str(lt)
        except Exception as e:
            raise ValueError(f"Could not parse layer_types: {layer_types}. {e}") from e
        return layer_types

    @validator("layer_types")
    def _nonempty_layer_types(cls, layer_types: List[str]) -> List[str]:
        """If layer_types is empty, raise a ValueError."""
        if not layer_types:
            raise ValueError("layer_types must not be empty")
        return layer_types

    @validator("layer_types")
    def _layer_types_unique(cls, layer_types: List[str]) -> List[str]:
        """Each layer type can be refered to at most once."""
        from collections import Counter

        c = Counter(LayerTypeConstraint.from_str(lt).layer_type for lt in layer_types)
        if any(c[lt] > 1 for lt in c):
            raise ValueError(f"Duplicate layer type in {layer_types}")
        return layer_types

    @validator("filename_extensions")
    def _coerce_common_glob_patterns(cls, exts: List[str]) -> List[str]:
        """If any of the listed extensions are common glob patterns, replace the
        list with one of all extensions.

        Coercions:
        1. File extensions beginning with '*' have their leading '*' removed.
        2. File extensions lacking a leading '.' have a leading '.' added.

        Rules:
        3. File extensions must start with '.' or '*.'
        4. File extensions must be at least two characters long.
        """

        exts = [f".{e.lstrip('*.')}" for e in exts]

        if any(len(e) < 2 for e in exts):
            raise ValueError(
                "Invalid file extension: Must have one character past the '.'"
            )
        return exts



import re

_package_name = "([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])"
_python_identifier = "([a-zA-Z_][a-zA-Z_0-9]*)"

# how do we deal with keywords ?
# do we try to validate ? Or do we just
# assume users won't try to create a command named
# `npe2_tester.False.if.for.in` ?
_identifier_plus_dash = "(?:[a-zA-Z_][a-zA-Z_0-9-]+)"
_dotted_name = f"(?:(?:{_identifier_plus_dash}\\.)*{_identifier_plus_dash})"
PACKAGE_NAME_PATTERN = re.compile(f"^{_package_name}$", re.IGNORECASE)
DOTTED_NAME_PATTERN = re.compile(_dotted_name)
DISPLAY_NAME_PATTERN = re.compile(r"^[^\W_][\w -~:.'\"]{1,88}[^\W_]$")
PYTHON_NAME_PATTERN = re.compile(f"^({_dotted_name}):({_dotted_name})$")
COMMAND_ID_PATTERN = re.compile(
    f"^(({_package_name}\\.)*{_python_identifier})$", re.IGNORECASE
)


def command_id(id: str) -> str:
    if id and not COMMAND_ID_PATTERN.match(id):
        raise ValueError(
            f"{id!r} is not a valid command id. It must begin with the package name "
            "followed by a period, then may can only contain alphanumeric "
            "characters and underscores."
        )
    return id


def package_name(name: str) -> str:
    """Assert that `name` is a valid package name in accordance with PEP-0508."""
    if name and not PACKAGE_NAME_PATTERN.match(name):
        raise ValueError(
            f"{name!r} is not a valid python package name. "
            "See https://peps.python.org/pep-0508/#names "
        )
    return name


def python_name(name: str) -> str:
    """Assert that `name` is a valid python name: e.g. `module.submodule:funcname`"""
    if name and not PYTHON_NAME_PATTERN.match(name):
        msg = (
            f"{name!r} is not a valid python_name.  A python_name must "
            "be of the form '{obj.__module__}:{obj.__qualname__}' (e.g. "
            "'my_package.a_module:some_function')."
        )
        if ".<locals>." in name:
            *_, a, b = name.split(".<locals>.")
            a = a.split(":")[-1]
            msg += (
                " Note: functions defined in local scopes are not yet supported. "
                f"Please move function {b!r} to the global scope of module {a!r}"
            )
        raise ValueError(msg)
    return name


def display_name(v: str) -> str:
    if not DISPLAY_NAME_PATTERN.match(v):
        raise ValueError(
            f"{v} is not a valid display_name. It must be 3-90 characters long, "
            "and must not begin or end with an underscore, white space, or "
            "non-word character."
        )
    return v


def icon_path(v: str) -> str:
    if not v:
        return ""
    if v.startswith("http"):
        if not v.startswith("https://"):
            raise ValueError(
                f"{v} is not a valid icon URL. It must start with 'https://'"
            )
        return v
    assert isinstance(v, str), f"{v} must be a string"
    return v



from importlib.metadata import PackageNotFoundError

import pytest

from npe2._inspection import _from_npe1
from npe2._inspection._from_npe1 import (
    convert_repository,
    get_top_module_path,
    manifest_from_npe1,
)


@pytest.mark.filterwarnings("ignore:The distutils package is deprecated")
@pytest.mark.filterwarnings("ignore:Found a multi-layer writer in")
@pytest.mark.parametrize("package", ["svg"])
def test_conversion(package):
    assert manifest_from_npe1(package)


@pytest.mark.filterwarnings("ignore:Failed to convert napari_provide_sample_data")
@pytest.mark.filterwarnings("ignore:Error converting function")
@pytest.mark.filterwarnings("ignore:Error converting dock widget")
def test_conversion_from_module(mock_npe1_pm, npe1_plugin_module):
    mf = manifest_from_npe1(module=npe1_plugin_module)
    assert isinstance(mf.dict(), dict)


def test_conversion_from_obj_with_locals(mock_npe1_pm):
    from napari_plugin_engine import napari_hook_implementation

    class MyPlugin:
        @staticmethod
        @napari_hook_implementation
        def napari_experimental_provide_function():
            def f(x: int):
                ...

            return [f]

    with pytest.warns(UserWarning) as record:
        mf = manifest_from_npe1(module=MyPlugin)
    msg = str(record[0].message)
    assert "functions defined in local scopes are not yet supported." in msg
    assert isinstance(mf.dict(), dict)


@pytest.mark.filterwarnings("ignore:Failed to convert napari_provide_sample_data")
@pytest.mark.filterwarnings("ignore:Error converting function")
@pytest.mark.filterwarnings("ignore:Error converting dock widget")
def test_conversion_from_package(npe1_repo, mock_npe1_pm_with_plugin):
    setup_cfg = npe1_repo / "setup.cfg"
    before = setup_cfg.read_text()
    convert_repository(npe1_repo, dry_run=True)
    assert setup_cfg.read_text() == before
    assert not (npe1_repo / "npe1_module" / "napari.yaml").exists()
    convert_repository(npe1_repo, dry_run=False)
    new_setup = setup_cfg.read_text()
    assert new_setup != before
    assert (
        "[options.entry_points]\n"
        "napari.manifest = \n	npe1-plugin = npe1_module:napari.yaml"
    ) in new_setup
    assert "[options.package_data]\nnpe1_module = napari.yaml" in new_setup
    assert (npe1_repo / "npe1_module" / "napari.yaml").is_file()

    with pytest.raises(ValueError) as e:
        convert_repository(npe1_repo)
    assert "Is this package already converted?" in str(e.value)


def _assert_expected_errors(record: pytest.WarningsRecorder):
    assert len(record) == 4
    msg = str(record[0].message)
    assert "Error converting dock widget [2] from 'npe1_module'" in msg
    msg = str(record[1].message)
    assert "Error converting function [1] from 'npe1_module'" in msg
    msg = str(record[2].message)
    assert "Failed to convert napari_provide_sample_data in 'npe1-plugin'" in msg
    assert "could not get resolvable python name" in msg
    msg = str(record[3].message)
    assert "Cannot auto-update setup.py, please edit setup.py as follows" in msg
    assert "npe1-plugin = npe1_module:napari.yaml" in msg


def test_conversion_from_package_setup_py(npe1_repo, mock_npe1_pm_with_plugin):
    (npe1_repo / "setup.cfg").unlink()
    (npe1_repo / "setup.py").write_text(
        """from setuptools import setup

NAME = 'npe1-plugin'
setup(
    name=NAME,
    entry_points={"napari.plugin": ["npe1-plugin = npe1_module"]}
)
"""
    )
    with pytest.warns(UserWarning) as record:
        convert_repository(npe1_repo)
    _assert_expected_errors(record)


def test_conversion_entry_point_string(npe1_repo, mock_npe1_pm_with_plugin):
    (npe1_repo / "setup.cfg").unlink()
    (npe1_repo / "setup.py").write_text(
        """from setuptools import setup

setup(
    name='npe1-plugin',
    entry_points={"napari.plugin": "npe1-plugin = npe1_module"}
)
"""
    )
    with pytest.warns(UserWarning) as record:
        convert_repository(npe1_repo)
    _assert_expected_errors(record)


def test_conversion_missing():
    with pytest.raises(ModuleNotFoundError), pytest.warns(UserWarning):
        manifest_from_npe1("does-not-exist-asdf6as987")


def test_conversion_package_is_not_a_plugin():
    with pytest.raises(PackageNotFoundError):
        manifest_from_npe1("pytest")


def test_get_top_module_path(mock_npe1_pm_with_plugin):
    get_top_module_path("npe1-plugin")


def test_python_name_local():
    def f():
        return lambda x: None

    with pytest.raises(ValueError) as e:
        _from_npe1._python_name(f())

    assert "functions defined in local scopes are not yet supported" in str(e.value)


def test_guess_fname_patterns():
    def get_reader1(path):
        if isinstance(path, str) and path.endswith((".tiff", ".tif")):
            return 1

    def get_reader2(path):
        if path.endswith(".xyz"):
            return 1

    assert _from_npe1._guess_fname_patterns(get_reader1) == ["*.tiff", "*.tif"]
    assert _from_npe1._guess_fname_patterns(get_reader2) == ["*.xyz"]



import sys
from concurrent.futures import ThreadPoolExecutor
from unittest.mock import patch

import pytest
from typer.testing import CliRunner

from npe2 import __version__
from npe2.cli import app, main
from npe2.manifest.schema import PluginManifest

runner = CliRunner()


@pytest.mark.parametrize("debug", ["--debug", ""])
@pytest.mark.parametrize("imports", ["--imports", "--no-imports"])
def test_cli_validate_ok(sample_path, debug, imports, monkeypatch):
    cmd = ["validate", str(sample_path / "my_plugin" / "napari.yaml"), imports]
    if debug:
        cmd += [debug]
    with monkeypatch.context() as m:
        m.setattr(sys, "path", [*sys.path, str(sample_path)])
        result = runner.invoke(app, cmd)
    assert "✔ Manifest for 'My Plugin' valid!" in result.stdout
    assert result.exit_code == 0


def test_cli_validate_invalid(tmp_path, capsys):
    (tmp_path / "manifest.yaml").write_text("name: hi??\n")
    cmd = ["validate", str(tmp_path / "manifest.yaml")]
    result = runner.invoke(app, cmd)
    assert "'hi??' is not a valid python package name." in result.stdout


def test_cli_validate_load_err(tmp_path):
    non_existent = str(tmp_path / "manifest.yaml")
    result = runner.invoke(app, ["validate", non_existent])
    assert result.exit_code == 1
    assert "🅇 Unexpected error in" in result.stdout
    assert "Could not find manifest for" in result.stdout

    result = runner.invoke(app, ["validate", non_existent, "--debug"])
    assert "Could not find manifest for" in result.stdout


@pytest.mark.parametrize("format", ["json", "yaml", "toml", "csv"])
@pytest.mark.parametrize("to_file", [True, False])
def test_cli_parse(sample_path, format, tmp_path, to_file):
    cmd = ["parse", str(sample_path / "my_plugin" / "napari.yaml")]
    if to_file:
        dest = tmp_path / f"output.{format}"
        cmd.extend(["-o", str(dest)])
    else:
        cmd.extend(["-f", format])

    result = runner.invoke(app, cmd)
    if format == "csv":
        assert result.exit_code
        return

    assert result.exit_code == 0
    if to_file:
        assert dest.exists()
        assert PluginManifest.from_file(dest)
    else:
        assert "my-plugin" in result.stdout  # just prints the yaml


@pytest.mark.parametrize("format", ["json", "yaml", "toml", "csv"])
@pytest.mark.parametrize("to_file", [True, False])
@pytest.mark.parametrize("include_meta", [True, False])
def test_cli_fetch(format, tmp_path, to_file, include_meta):
    cmd = ["fetch", "napari-omero"]
    if to_file:
        dest = tmp_path / f"output.{format}"
        cmd.extend(["-o", str(dest)])
    else:
        cmd.extend(["-f", format])
    if include_meta:
        cmd.extend(["--include-package-meta", "--indent=2"])

    result = runner.invoke(app, cmd)
    if format == "csv":
        assert result.exit_code
        return

    assert result.exit_code == 0
    if to_file:
        assert dest.exists()
        assert PluginManifest.from_file(dest)
    else:
        assert "napari-omero" in result.stdout  # just prints the yaml
        if include_meta:
            assert "package_metadata" in result.stdout


def test_cli_fetch_all(tmp_path, monkeypatch):
    dest = tmp_path / "output"
    with patch("npe2._inspection._fetch.get_pypi_plugins") as mock_hub:
        mock_hub.return_value = {"a": "0.1.0", "b": "0.2.0", "c": "0.3.0"}
        with patch("npe2._inspection._fetch.ProcessPoolExecutor", ThreadPoolExecutor):
            cmd = ["fetch", "--all", "-o", str(dest)]
            monkeypatch.setattr(sys, "argv", cmd)
            result = runner.invoke(app, cmd)

    mock_hub.assert_called_once()
    assert result.exit_code == 0
    assert dest.exists()
    assert (dest / "errors.json").exists()


@pytest.mark.filterwarnings("default:Failed to convert")
def test_cli_convert_repo(npe1_repo, mock_npe1_pm_with_plugin):
    result = runner.invoke(app, ["convert", str(npe1_repo)])
    assert result.exit_code == 0
    assert "✔  Conversion complete!" in result.stdout


@pytest.mark.filterwarnings("default:Failed to convert")
def test_cli_convert_repo_dry_run(npe1_repo, mock_npe1_pm_with_plugin):
    result = runner.invoke(app, ["convert", str(npe1_repo), "-n"])
    assert result.exit_code == 0
    assert "# Manifest would be written to" in result.stdout


@pytest.mark.filterwarnings("ignore:The distutils package is deprecated")
@pytest.mark.filterwarnings("default:Failed to convert napari_get_writer")
def test_cli_convert_svg():
    result = runner.invoke(app, ["convert", "napari-svg"])
    assert "Some issues occured:" in result.stdout
    assert "Found a multi-layer writer in 'napari-svg'" in result.stdout
    assert result.exit_code == 0


def test_cli_convert_repo_fails(npe1_repo, mock_npe1_pm_with_plugin):
    (npe1_repo / "setup.cfg").unlink()
    result = runner.invoke(app, ["convert", str(npe1_repo)])
    assert result.exit_code == 1
    assert "Could not detect first gen napari plugin package" in result.stdout


@pytest.mark.filterwarnings("default:Failed to convert")
def test_cli_convert_package_name(npe1_repo, mock_npe1_pm_with_plugin):
    result = runner.invoke(app, ["convert", "npe1-plugin"])
    assert "name: npe1-plugin" in result.stdout  # just prints the yaml
    assert result.exit_code == 0


def test_cli_main(monkeypatch, sample_path):
    cmd = ["npe2", "validate", str(sample_path / "my_plugin" / "napari.yaml")]
    monkeypatch.setattr(sys, "argv", cmd)
    with pytest.raises(SystemExit) as e:
        main()
    assert e.value.code == 0


def test_cli_cache_list_empty():
    result = runner.invoke(app, ["cache", "--list"])
    assert "Nothing cached" in result.stdout
    assert result.exit_code == 0


def test_cli_cache_list_full(uses_npe1_plugin, mock_cache):
    (mock_cache / "npe1-plugin.yaml").write_text("name: npe1-plugin\n")
    result = runner.invoke(app, ["cache", "--list"])
    assert result.stdout == "npe1-plugin: 0.1.0\n"
    assert result.exit_code == 0


def test_cli_cache_list_named(uses_npe1_plugin, mock_cache):
    (mock_cache / "npe1-plugin.yaml").write_text("name: npe1-plugin\n")
    result = runner.invoke(app, ["cache", "--list", "not-a-plugin"])
    assert result.stdout == "Nothing cached for plugins: not-a-plugin\n"
    assert result.exit_code == 0


def test_cli_cache_clear_empty():
    result = runner.invoke(app, ["cache", "--clear"])
    assert "Nothing to clear" in result.stdout
    assert result.exit_code == 0


def test_cli_cache_clear_full(mock_cache):
    (mock_cache / "npe1-plugin.yaml").write_text("name: npe1-plugin\n")
    result = runner.invoke(app, ["cache", "--clear"])
    assert "Cleared these files from cache" in result.stdout
    assert "- npe1-plugin.yaml" in result.stdout
    assert result.exit_code == 0


def test_cli_cache_clear_named(mock_cache):
    (mock_cache / "npe1-plugin.yaml").write_text("name: npe1-plugin\n")
    result = runner.invoke(app, ["cache", "--clear", "not-a-plugin"])
    assert result.stdout == "Nothing to clear for plugins: not-a-plugin\n"
    assert result.exit_code == 0


@pytest.mark.parametrize("format", ["table", "compact", "yaml", "json"])
@pytest.mark.parametrize("fields", [None, "name,version,author"])
def test_cli_list(format, fields, uses_npe1_plugin):
    result = runner.invoke(app, ["list", "-f", format, "--fields", fields])
    assert result.exit_code == 0
    assert "npe1-plugin" in result.output
    if fields and "author" in fields and format != "compact":
        assert "author" in result.output.lower()
    else:
        assert "author" not in result.output.lower()


def test_cli_list_sort(uses_npe1_plugin):
    result = runner.invoke(app, ["list", "--sort", "version"])
    assert result.exit_code == 0

    result = runner.invoke(app, ["list", "--sort", "7"])
    assert result.exit_code
    assert "Invalid sort value '7'" in result.output

    result = runner.invoke(app, ["list", "--sort", "notaname"])
    assert result.exit_code
    assert "Invalid sort value 'notaname'" in result.output


def test_cli_version():
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert __version__ in result.output


def test_compile(compiled_plugin_dir):
    result = runner.invoke(app, ["compile", str(compiled_plugin_dir)])
    assert "id: my_compiled_plugin.my-plugin.generate_random_data" in result.output



import ast
import inspect
from abc import ABC, abstractmethod
from importlib.metadata import Distribution
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, DefaultDict, Dict, List, Tuple, Type, Union

from npe2.manifest import contributions

if TYPE_CHECKING:
    from pydantic import BaseModel


CONTRIB_MAP: Dict[str, Tuple[Type["BaseModel"], str]] = {
    "writer": (contributions.WriterContribution, "writers"),
    "reader": (contributions.ReaderContribution, "readers"),
    "sample_data_generator": (contributions.SampleDataGenerator, "sample_data"),
    "widget": (contributions.WidgetContribution, "widgets"),
}


class _DecoratorVisitor(ast.NodeVisitor, ABC):
    """Visitor that finds decorators that match something.

    1. Visit all `Import` and `ImportFrom` nodes in the module, storing their import
       names in `_names` so we can look them up later.  For example, if the module
       had the line `from npe2 import implements as impls` at the top, then the
       `visit_ImportFrom` method would add the entry:
       `self._names['impls'] = 'npe2.implements'`
       This way, we know that an `@impls` found later in the module is referring to
       `npe2.implements`.
    2. Visit all `FunctionDef` and `ClassDef` nodes in the module and check their
       decorators with `_find_decorators`
    3. In `_find_decorators` we check to see if the name of any of the decorators
       resolves to the something from self._match (i.e., if it's being decorated
       with something from `npe2.implements`).  If it is, then we call
       `_process_decorated`

    implement `_process_decorated` in subclasses
    """

    def __init__(self, module_name: str, match: str) -> None:
        self.module_name = module_name
        self._match = match
        self._names: Dict[str, str] = {}

    def visit_Import(self, node: ast.Import) -> Any:
        # https://docs.python.org/3/library/ast.html#ast.Import
        for alias in node.names:
            self._names[alias.asname or alias.name] = alias.name
        return super().generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> Any:
        # https://docs.python.org/3/library/ast.html#ast.ImportFrom
        module = node.module
        if node.level > 0:
            root = self.module_name.rsplit(".", node.level)[0]
            module = f"{root}.{module}"
        for alias in node.names:
            self._names[alias.asname or alias.name] = f"{module}.{alias.name}"
        return super().generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Any:
        # https://docs.python.org/3/library/ast.html#ast.FunctionDef
        self._find_decorators(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> Any:
        self._find_decorators(node)

    def _find_decorators(self, node: Union[ast.ClassDef, ast.FunctionDef]):
        # for each in the decorator list ...
        for call in node.decorator_list:
            # https://docs.python.org/3/library/ast.html#ast.Call
            if isinstance(call, ast.Name) and self._names.get(call.id, "").startswith(
                self._match
            ):
                self._process_decorated(call.id, node, {})

            if not isinstance(call, ast.Call):
                continue
            # if the function is an attribute ...
            # (e.g in `@npe2.implements.reader`, `reader` is an attribute of
            # implements, which is an attribute of npe2)
            if isinstance(call.func, ast.Attribute):
                # then go up the chain of attributes until we get to a root Name
                val = call.func.value
                _names = []
                while isinstance(val, ast.Attribute):
                    _names.append(val.attr)
                    val = val.value
                if isinstance(val, ast.Name):
                    _names.append(val.id)
                    # finally, we can build the fully resolved call name of the
                    # decorator (e.g. `@npe2.implements`, or `@implements`)
                    call_name = ".".join(reversed(_names))
                    # we then check the `_names` we gathered during imports to resolve
                    # these call names to their fully qualified names (e.g. `implements`
                    # would resolve to `npe2.implements` if the name `implements` was
                    # imported from `npe2`.)
                    # If the name resolves to whatever `self._match` is,
                    # then we have a hit! process the function.
                    # In the case of an NPE2 module visitor, the name we're trying to
                    # match will be `npe2.implements` or `implements`.
                    # In the case of a npe1 visitor, the name we're trying to match
                    # will be `napari_plugin_engine.napari_hook_implementation`.
                    if self._names.get(call_name) == self._match:
                        kwargs = self._keywords_to_kwargs(call.keywords)
                        self._process_decorated(call.func.attr, node, kwargs)
            elif isinstance(call.func, ast.Name):
                # if the function is just a direct name (e.g. `@reader`)
                # then we can just see if the name points to something imported from
                # this module.
                if self._names.get(call.func.id, "").startswith(self._match):
                    kwargs = self._keywords_to_kwargs(call.keywords)
                    self._process_decorated(call.func.id, node, kwargs)
        return super().generic_visit(node)

    def _keywords_to_kwargs(self, keywords: List[ast.keyword]) -> Dict[str, Any]:
        return {str(k.arg): ast.literal_eval(k.value) for k in keywords}

    @abstractmethod
    def _process_decorated(
        self,
        decorator_name: str,
        node: Union[ast.ClassDef, ast.FunctionDef],
        decorator_kwargs: dict,
    ):
        """Process a decorated function.

        This is a hook for subclasses to do something with the decorated function.
        """


class NPE2PluginModuleVisitor(_DecoratorVisitor):
    """AST visitor to find all contributions in an npe2 module.

    This visitor will find all the contributions (things decorated with
    `@npe2.implements`) in a module and store them in `contribution_points`.

    It works as follows:

    See how decorators are found in the docstring for `_DecoratorVisitor`.
    when a decorator is found that matches `npe2.implements` (e.g. `@implements.reader`)

    4. `_store_contrib` first calls `_store_command` which does the job of storing the
       fully-resolved `python_name` for the function being decorated, and creates a
       CommandContribution. `_store_contrib` will then create the appropriate
       contribution type (e.g. `npe2.implements.reader` will instantiate a
       `ReaderContribution`), set its `command` entry to the `id` of the just-created
       `CommandContribution`, then store it in `contribution_points`.
    5. When the visitor is finished, we can create an instance of `ContributionPoints`
       using `ContributionPoints(**visitor.contribution_points)`, then add it to a
       manifest.
    """

    def __init__(
        self, plugin_name: str, module_name: str, match: str = "npe2.implements"
    ) -> None:
        super().__init__(module_name, match)
        self.plugin_name = plugin_name
        self.contribution_points: Dict[str, List[dict]] = {}

    def _process_decorated(
        self,
        decorator_name: str,
        node: Union[ast.ClassDef, ast.FunctionDef],
        decorator_kwargs: dict,
    ):
        self._store_contrib(decorator_name, node.name, decorator_kwargs)

    def _store_contrib(self, contrib_type: str, name: str, kwargs: Dict[str, Any]):
        from npe2.implements import CHECK_ARGS_PARAM  # circ import

        kwargs.pop(CHECK_ARGS_PARAM, None)
        ContribClass, contrib_name = CONTRIB_MAP[contrib_type]
        contrib = ContribClass(**self._store_command(name, kwargs))
        existing: List[dict] = self.contribution_points.setdefault(contrib_name, [])
        existing.append(contrib.dict(exclude_unset=True))

    def _store_command(self, name: str, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        cmd_params = inspect.signature(contributions.CommandContribution).parameters

        cmd_kwargs = {k: kwargs.pop(k) for k in list(kwargs) if k in cmd_params}
        cmd_kwargs["python_name"] = self._qualified_pyname(name)
        cmd = contributions.CommandContribution(**cmd_kwargs)
        if cmd.id.startswith(self.plugin_name):
            n = len(self.plugin_name)
            cmd.id = cmd.id[n:]
        cmd.id = f"{self.plugin_name}.{cmd.id.lstrip('.')}"
        cmd_contribs: List[dict] = self.contribution_points.setdefault("commands", [])
        cmd_contribs.append(cmd.dict(exclude_unset=True))
        kwargs["command"] = cmd.id
        return kwargs

    def _qualified_pyname(self, obj_name: str) -> str:
        return f"{self.module_name}:{obj_name}"


class NPE1PluginModuleVisitor(_DecoratorVisitor):
    """AST visitor to find all contributions in an npe1 module."""

    def __init__(self, plugin_name: str, module_name: str) -> None:
        super().__init__(module_name, "napari_plugin_engine.napari_hook_implementation")
        self.plugin_name = plugin_name
        self.contribution_points: DefaultDict[str, list] = DefaultDict(list)

    def _process_decorated(
        self,
        decorator_name: str,
        node: Union[ast.ClassDef, ast.FunctionDef],
        decorator_kwargs: Dict[str, Any],
    ):
        self.generic_visit(node)  # do this to process any imports in the function
        hookname = decorator_kwargs.get("specname", node.name)
        getattr(self, hookname)(node)

    def _add_command(self, node: ast.FunctionDef) -> contributions.CommandContribution:
        cmd_id = f"{self.plugin_name}.{node.name}"
        py_name = f"{self.module_name}:{node.name}"
        cmd_contrib = contributions.CommandContribution(
            id=cmd_id, python_name=py_name, title=node.name
        )
        self.contribution_points["commands"].append(cmd_contrib)
        return cmd_contrib

    def napari_get_reader(self, node: ast.FunctionDef):
        cmd = self._add_command(node)
        rdr_contrib = contributions.ReaderContribution(
            command=cmd.id, filename_patterns=["*"], accepts_directories=True
        )
        self.contribution_points["readers"].append(rdr_contrib)

    def napari_get_writer(self, node: ast.FunctionDef):
        # we can't convert this to an npe2 command contribution
        pass  # pragma: no cover

    def napari_write_image(self, node: ast.FunctionDef):
        self._parse_writer(node, "image")

    def napari_write_labels(self, node: ast.FunctionDef):
        self._parse_writer(node, "labels")  # pragma: no cover

    def napari_write_points(self, node: ast.FunctionDef):
        self._parse_writer(node, "points")  # pragma: no cover

    def napari_write_shapes(self, node: ast.FunctionDef):
        self._parse_writer(node, "shapes")  # pragma: no cover

    def napari_write_vectors(self, node: ast.FunctionDef):
        self._parse_writer(node, "vectors")  # pragma: no cover

    def _parse_writer(self, node, layer_type: str):
        cmd = self._add_command(node)
        wrt_contrib = contributions.WriterContribution(
            command=cmd.id, layer_types=[layer_type], display_name=layer_type
        )
        self.contribution_points["writers"].append(wrt_contrib)

    def napari_provide_sample_data(self, node: ast.FunctionDef):
        from npe2.manifest.utils import safe_key

        return_ = next(n for n in node.body if isinstance(n, ast.Return))
        if not isinstance(return_.value, ast.Dict):
            raise TypeError(  # pragma: no cover
                f"napari_provide_sample_data must return a dict, not {type(return_)}"
            )

        contrib: contributions.SampleDataContribution
        for key, val in zip(return_.value.keys, return_.value.values):
            if isinstance(val, ast.Dict):
                raise NotImplementedError("npe1 sample dicts-of-dicts not supported")

            assert isinstance(key, ast.Constant)
            display_name = key.value
            key = safe_key(display_name)  # type: ignore

            # sample should now either be a callable, or a string
            if isinstance(val, ast.Name):
                cmd_id = f"{self.plugin_name}.{val.id}"
                py_name = f"{self.module_name}:{val.id}"
                cmd_contrib = contributions.CommandContribution(
                    id=cmd_id, python_name=py_name, title=val.id
                )
                self.contribution_points["commands"].append(cmd_contrib)
                contrib = contributions.SampleDataGenerator(
                    command=cmd_id, key=key, display_name=display_name
                )
            else:
                uri = "__dynamic__"  # TODO: make this a real uri
                contrib = contributions.SampleDataURI(
                    key=key, display_name=display_name, uri=uri
                )

            self.contribution_points["sample_data"].append(contrib)

    def napari_experimental_provide_function(self, node: ast.FunctionDef):
        return_ = next(n for n in node.body if isinstance(n, ast.Return))

        items = (
            list(return_.value.elts)
            if isinstance(return_.value, ast.List)
            else [return_.value]  # type: ignore
        )

        for item in items:
            if not isinstance(item, ast.Name):
                raise NotImplementedError(  # pragma: no cover
                    "provide function got non-name"
                )

            py_name = self._names.get(item.id)
            py_name = (
                ":".join(py_name.rsplit(".", 1))
                if py_name
                else f"{self.module_name}:{node.name}"
            )

            cmd_id = f"{self.plugin_name}.{node.name}"
            cmd_contrib = contributions.CommandContribution(
                id=cmd_id, python_name=py_name, title=item.id
            )

            wdg_contrib = contributions.WidgetContribution(
                command=cmd_id, display_name=item.id, autogenerate=True
            )
            self.contribution_points["commands"].append(cmd_contrib)
            self.contribution_points["widgets"].append(wdg_contrib)

    def napari_experimental_provide_dock_widget(self, node: ast.FunctionDef):
        return_ = next(n for n in node.body if isinstance(n, ast.Return))
        items = (
            list(return_.value.elts)
            if isinstance(return_.value, ast.List)
            else [return_.value]  # type: ignore
        )

        for item in items:
            wdg_creator = item.elts[0] if isinstance(item, ast.Tuple) else item
            if isinstance(wdg_creator, ast.Name):
                # eg `SegmentationWidget`
                obj_name = wdg_creator.id
                if py_name := self._names.get(wdg_creator.id):
                    py_name = ":".join(py_name.rsplit(".", 1))
                else:
                    py_name = f"{self.module_name}:{obj_name}"
            elif isinstance(wdg_creator, ast.Attribute):
                # eg `measurement.analyze_points_layer`
                py_name = wdg_creator.attr
                tmp = wdg_creator
                assert isinstance(tmp.value, ast.Name)
                py_name = f"{self._names[tmp.value.id]}.{py_name}"
                py_name = ":".join(py_name.rsplit(".", 1))
                obj_name = tmp.value.id
            else:
                raise TypeError(  # pragma: no cover
                    f"Unexpected widget creator type: {type(wdg_creator)}"
                )

            cmd_id = f"{self.plugin_name}.{obj_name}"
            cmd_contrib = contributions.CommandContribution(
                id=cmd_id, python_name=py_name, title=obj_name
            )
            wdg_contrib = contributions.WidgetContribution(
                command=cmd_id, display_name=obj_name
            )
            self.contribution_points["commands"].append(cmd_contrib)
            self.contribution_points["widgets"].append(wdg_contrib)


def find_npe2_module_contributions(
    path: Union[ModuleType, str, Path], plugin_name: str, module_name: str = ""
) -> contributions.ContributionPoints:
    """Visit an npe2 module and extract contribution points.

    Parameters
    ----------
    path : Union[ModuleType, str, Path]
        Either a path to a Python module, a module object, or a string
    plugin_name : str
        Name of the plugin
    module_name : str
        Module name, by default ""

    Returns
    -------
    ContributionPoints
        ContributionPoints discovered in the module.
    """
    if isinstance(path, ModuleType):
        assert path.__file__
        assert path.__name__
        module_name = path.__name__
        path = path.__file__

    visitor = NPE2PluginModuleVisitor(plugin_name, module_name=module_name)
    visitor.visit(ast.parse(Path(path).read_text()))
    if "commands" in visitor.contribution_points:
        compress = {tuple(i.items()) for i in visitor.contribution_points["commands"]}
        visitor.contribution_points["commands"] = [dict(i) for i in compress]
    return contributions.ContributionPoints(**visitor.contribution_points)


def find_npe1_module_contributions(
    dist: Distribution, module_name: str = ""
) -> contributions.ContributionPoints:
    """Statically visit an npe1 module and extract contribution points.

    Parameters
    ----------
    dist: Distribution
        A distribution object representing an npe1 plugin to be visited.
    module_name : str
        Module name, by default ""

    Returns
    -------
    ContributionPoints
        ContributionPoints discovered in the module.
    """
    plugin_name = dist.metadata["Name"]
    file = _locate_module_in_dist(dist, module_name)
    visitor = NPE1PluginModuleVisitor(plugin_name, module_name=module_name)
    visitor.visit(ast.parse(Path(file).read_text()))

    # now check all of the modules that were imported by `module_name` to see
    # if any of those had npe1 decorated functions.
    # NOTE: we're only going 1 level deep here...
    for name, target in visitor._names.items():
        if not name.startswith("_"):
            target_module = target.rsplit(".", 1)[0]
            try:
                file = _locate_module_in_dist(dist, target_module)
            except FileNotFoundError:
                # if the imported module is not in the same distribution
                # just skip it.
                continue
            # NOTE: technically, this time we should restrict the allowable names
            # to those that are imported from the original module_name ...
            # but that's probably overkill
            v2 = NPE1PluginModuleVisitor(plugin_name, target_module)
            v2.visit(ast.parse(file.read_text()))
            visitor.contribution_points.update(v2.contribution_points)

    return contributions.ContributionPoints(**visitor.contribution_points)


def _locate_module_in_dist(dist: Distribution, module_name: str) -> Path:
    root = dist.locate_file(module_name.replace(".", "/"))
    assert isinstance(root, Path)

    if not (file := root / "__init__.py").exists():
        if not (file := root.with_suffix(".py")).exists():
            raise FileNotFoundError(
                f"Could not find {module_name!r} in "
                f"in distribution for {dist.metadata['Name']}"
            )
    return file



from typing import TYPE_CHECKING, Any, Dict, List, Tuple

from pydantic import BaseModel

from npe2 import PluginContext
from npe2.types import PathOrPaths

if TYPE_CHECKING:
    import napari.types


def activate(context: PluginContext):
    @context.register_command("my_plugin.hello_world")
    def _hello():
        ...

    context.register_command("my_plugin.another_command", lambda: print("yo!"))


def deactivate(context: PluginContext):
    """just here for tests"""


def get_reader(path: PathOrPaths):
    if isinstance(path, list):

        def read(path):
            assert isinstance(path, list)
            return [(None,)]

        return read
    assert isinstance(path, str)  # please mypy.
    if path.endswith(".fzzy"):

        def read(path):
            assert isinstance(path, str)
            return [(None,)]

        return read
    else:
        raise ValueError("Test plugin should not receive unknown data")


def url_reader(path: str):
    if path.startswith("http"):

        def read(path):
            return [(None,)]

        return read


def writer_function(path: str, layer_data: List[Tuple[Any, Dict, str]]) -> List[str]:
    class Arg(BaseModel):
        data: Any
        meta: Dict
        layer_type: str

    for e in layer_data:
        Arg(data=e[0], meta=e[1], layer_type=e[2])

    return [path]


def writer_function_single(path: str, layer_data: Any, meta: Dict) -> List[str]:
    class Arg(BaseModel):
        data: Any
        meta: Dict

    Arg(data=layer_data, meta=meta)

    return [path]


class SomeWidget:
    ...


def random_data():
    import numpy as np

    return [(np.random.rand(10, 10))]


def make_widget_from_function(image: "napari.types.ImageData", threshold: int):
    ...



from typing import Any, Callable, List, TypeVar

from pydantic import BaseModel as BaseModel

from .manifest import PluginManifest as PluginManifest
from .manifest import contributions as contributions

T = TypeVar("T", bound=Callable[..., Any])

CHECK_ARGS_PARAM: str

def reader(
    *,
    id: str,
    title: str,
    filename_patterns: List[str],
    accepts_directories: bool = False,
    ensure_args_valid: bool = False,
) -> Callable[[T], T]:
    """Mark a function as a reader contribution"""

def writer(
    *,
    id: str,
    title: str,
    layer_types: List[str],
    filename_extensions: List[str] = [],  # noqa: B006
    display_name: str = "",
    ensure_args_valid: bool = False,
) -> Callable[[T], T]:
    """Mark function as a writer contribution"""

def widget(
    *,
    id: str,
    title: str,
    display_name: str,
    autogenerate: bool = False,
    ensure_args_valid: bool = False,
) -> Callable[[T], T]:
    """Mark a function as a widget contribution"""

def sample_data_generator(
    *,
    id: str,
    title: str,
    key: str,
    display_name: str,
    ensure_args_valid: bool = False,
) -> Callable[[T], T]:
    """Mark a function as a sample data generator contribution"""

def on_activate(func):
    """Mark a function to be called when a plugin is activated."""

def on_deactivate(func):
    """Mark a function to be called when a plugin is deactivated."""



from importlib import metadata
from pathlib import Path
from unittest.mock import patch

import pytest
from pydantic import ValidationError

from npe2 import PluginManifest
from npe2.manifest import PackageMetadata
from npe2.manifest.schema import ENTRY_POINT

SAMPLE_PLUGIN_NAME = "my-plugin"
SAMPLE_MODULE_NAME = "my_plugin"


def test_sample_plugin_valid(sample_manifest):
    assert sample_manifest


def test_discover_empty():
    # sanity check to make sure sample_plugin must be in path to be discovered
    results = PluginManifest.discover()
    manifests = [result.manifest.name for result in results if result.manifest]
    assert SAMPLE_PLUGIN_NAME not in manifests


def test_schema():
    assert isinstance(PluginManifest.schema_json(), str)

    dschema = PluginManifest.schema()
    assert isinstance(dschema, dict)
    assert "name" in dschema["properties"]


def test_discover(uses_sample_plugin):
    discover_results = list(PluginManifest.discover())
    assert len(discover_results) == 1
    [(manifest, distribution, error)] = discover_results
    assert manifest and manifest.name == SAMPLE_PLUGIN_NAME
    assert distribution
    entrypoint = tuple(distribution.entry_points)[0]
    assert entrypoint and entrypoint.group == "napari.manifest" == ENTRY_POINT
    assert entrypoint.value == f"{SAMPLE_MODULE_NAME}:napari.yaml"
    assert error is None


def test_discover_errors(tmp_path: Path):
    """testing various discovery errors"""
    # package with proper `napari.manifest` entry_point, but invalid pointer to
    # a manifest should yield an error in results
    a = tmp_path / "a"
    a.mkdir()
    a_ep = a / "entry_points.txt"
    bad_value = "asdfsad:blahblahblah.yaml"
    a_ep.write_text(f"[napari.manifest]\n{SAMPLE_PLUGIN_NAME} = {bad_value}")

    # package with proper `napari.manifest` entry_point, but invalid manifest
    b = tmp_path / "b"
    b.mkdir()
    b_ep = b / "entry_points.txt"
    b_ep.write_text("[napari.manifest]\nsome_plugin = module:napari.yaml")
    module = tmp_path / "module"
    module.mkdir()
    (module / "napari.yaml").write_text("name: hi??")

    # a regular package, with out napari.manifest entry_point should just be skipped
    c = tmp_path / "c"
    c.mkdir()
    c_ep = c / "entry_points.txt"
    c_ep.write_text("[console.scripts]\nsomething = something")

    dists = [
        metadata.PathDistribution(a),
        metadata.PathDistribution(b),
        metadata.PathDistribution(c),
    ]

    with patch.object(metadata, "distributions", return_value=dists):
        discover_results = list(PluginManifest.discover(paths=[tmp_path]))

    assert len(discover_results) == 2
    res_a, res_b = discover_results
    assert res_a.manifest is None
    assert res_a.distribution
    assert tuple(res_a.distribution.entry_points)[0].value == bad_value
    assert "Cannot find module 'asdfsad'" in str(res_a.error)

    assert res_b.manifest is None
    assert res_b.distribution
    assert tuple(res_b.distribution.entry_points)[0].value == "module:napari.yaml"
    assert isinstance(res_b.error, ValidationError)


def test_package_meta(uses_sample_plugin):
    direct_meta = PackageMetadata.for_package(SAMPLE_PLUGIN_NAME)
    assert direct_meta.name == SAMPLE_PLUGIN_NAME
    assert direct_meta.version == "1.2.3"
    discover_results = list(PluginManifest.discover())
    [(manifest, *_)] = discover_results
    assert manifest
    assert manifest.package_metadata == direct_meta

    assert manifest.author == direct_meta.author == "The Black Knight"
    assert manifest.description == direct_meta.summary == "My napari plugin"
    assert manifest.package_version == direct_meta.version == "1.2.3"
    assert manifest.license == direct_meta.license == "BSD-3"


def test_all_package_meta():
    """make sure PackageMetadata works for whatever packages are in the environment.

    just a brute force way to get a little more validation coverage
    """
    for d in metadata.distributions():
        assert PackageMetadata.from_dist_metadata(d.metadata)


@pytest.mark.parametrize("format", ["toml", "json", "yaml", "pyproject"])
def test_export_round_trip(sample_manifest, tmp_path, format):
    """Test that an exported manifest can be round-tripped."""
    if format == "pyproject":
        out_file = tmp_path / "pyproject.toml"
        out_file.write_text(sample_manifest.toml(pyproject=True))
    else:
        out_file = tmp_path / f"napari.{format}"
        out_file.write_text(getattr(sample_manifest, format)())
    assert sample_manifest == PluginManifest.from_file(out_file)


def test_from_distribution(uses_sample_plugin):
    mf = PluginManifest.from_distribution(SAMPLE_PLUGIN_NAME)
    assert mf.name == SAMPLE_PLUGIN_NAME
    assert mf.package_metadata == PackageMetadata.for_package(SAMPLE_PLUGIN_NAME)

    with pytest.raises(metadata.PackageNotFoundError):
        _ = PluginManifest.from_distribution("not-an-installed-package")

    with pytest.raises(ValueError) as e:
        # valid package, but doesn't have a manifest
        _ = PluginManifest.from_distribution("pytest")
    assert "exists but does not provide a napari manifest" in str(e.value)


def test_from_package_name_err():
    with pytest.raises(ValueError) as e:
        PluginManifest._from_package_or_name("nonsense")
    assert "Could not find manifest for 'nonsense'" in str(e.value)


def test_dotted_name_with_command():
    with pytest.raises(ValidationError, match="must start with the current package"):
        PluginManifest(
            name="plugin.plugin-sample",
            contributions={"commands": [{"id": "plugin.command", "title": "Sample"}]},
        )
    with pytest.raises(ValidationError, match="must begin with the package name"):
        PluginManifest(
            name="plugin.plugin-sample",
            contributions={
                "commands": [{"id": "plugin.plugin-samplecommand", "title": "Sample"}]
            },
        )

    PluginManifest(
        name="plugin.plugin-sample",
        contributions={
            "commands": [{"id": "plugin.plugin-sample.command", "title": "Sample"}]
        },
    )


def test_visibility():
    mf = PluginManifest(name="myplugin")
    assert mf.is_visible

    mf = PluginManifest(name="myplugin", visibility="hidden")
    assert not mf.is_visible

    with pytest.raises(ValidationError):
        mf = PluginManifest(name="myplugin", visibility="other")


def test_icon():
    PluginManifest(name="myplugin", icon="my_plugin:myicon.png")


def test_dotted_plugin_name():
    """Test that"""
    name = "some.namespaced.plugin"
    cmd_id = f"{name}.frame_rate_widget"
    mf = PluginManifest(
        name=name,
        contributions={
            "commands": [
                {
                    "id": cmd_id,
                    "title": "open my widget",
                }
            ],
            "widgets": [
                {
                    "command": cmd_id,
                    "display_name": "Plot frame rate",
                }
            ],
        },
    )
    assert mf.contributions.widgets
    assert mf.contributions.widgets[0].plugin_name == name



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Type, Union

from pydantic import BaseModel, Field, PrivateAttr, conlist, root_validator, validator

if TYPE_CHECKING:
    from jsonschema.exceptions import ValidationError
    from jsonschema.protocols import Validator
else:
    try:
        from jsonschema.exceptions import ValidationError
    except ImportError:
        ValidationError = Exception

__all__ = [
    "ValidationError",
    "Draft04JsonSchema",
    "Draft06JsonSchema",
    "Draft07JsonSchema",
]

JsonType = Literal["array", "boolean", "integer", "null", "number", "object", "string"]
JsonTypeArray = conlist(JsonType, min_items=True, unique_items=True)
StringArrayMin1 = conlist(str, unique_items=True, min_items=1)
StringArray = conlist(str, unique_items=True)

PY_NAME_TO_JSON_NAME = {
    "list": "array",
    "bool": "boolean",
    "int": "integer",
    "float": "number",
    "dict": "object",
    "str": "string",
    "NoneType": "null",
    "None": "null",
}


def _to_json_type(type_: Union[str, Type]) -> JsonType:
    if isinstance(type_, type):
        type_ = type_.__name__
    type_ = str(type_).lower()
    return PY_NAME_TO_JSON_NAME.get(type_, type_)  # type: ignore # (validated later)


def _coerce_type_name(v):
    """Coerce python type names to json schema type names."""
    if isinstance(v, list):
        return [_to_json_type(t) for t in v]
    return _to_json_type(v)


def _to_camel(string: str) -> str:
    words = string.split("_")
    return words[0] + "".join(w.capitalize() for w in words[1:])


_CONSTRAINT_FIELDS = {
    "exclusive_minimum",
    "minimum",
    "exclusive_maximum",
    "maximum",
    "multiple_of",
    "min_items",
    "max_items",
    "min_length",
    "max_length",
    "pattern",
}

_python_equivalent: Dict[Optional[str], Type] = {
    "array": list,
    "boolean": bool,
    "integer": int,
    "null": type(None),
    "number": float,
    "object": dict,
    "string": str,
    None: object,
}


class _JsonSchemaBase(BaseModel):
    class Config:
        alias_generator = _to_camel
        allow_population_by_field_name = True

    # underscore here to avoid name collision with pydantic's `schema` method
    schema_: Optional[str] = Field(None, alias="$schema")
    title: Optional[str] = Field(None)
    description: Optional[str] = Field(None)
    default: Any = Field(None)
    multiple_of: Optional[float] = Field(None, ge=0)
    maximum: Optional[float] = Field(None)
    minimum: Optional[float] = Field(None)
    max_length: Optional[int] = Field(None, ge=0)
    min_length: Optional[int] = Field(0, ge=0)
    # could be Pattern. but it's easier to work with as str
    pattern: Optional[str] = Field(None)
    max_items: Optional[int] = Field(None, ge=0)
    min_items: Optional[int] = Field(0, ge=0)
    unique_items: bool = Field(False)
    max_properties: Optional[int] = Field(None, ge=0)
    min_properties: Optional[int] = Field(0, ge=0)
    enum: Optional[conlist(Any, min_items=1, unique_items=True)] = Field(None)  # type: ignore  # noqa
    type: Union[JsonType, JsonTypeArray] = Field(None)  # type: ignore
    format: Optional[str] = Field(None)

    _json_validator: Type[Validator] = PrivateAttr()

    # these will be redefined in subclasses with specific subschema types
    # just here for type-checking in the methods of this base class
    if TYPE_CHECKING:
        items: Any
        properties: Any
        all_of: Any
        any_of: Any
        one_of: Any

    _coerce_type_name = validator("type", pre=True, allow_reuse=True)(_coerce_type_name)

    @root_validator(pre=True)
    def _validate_root(cls, values: Dict[str, Any]) -> Any:
        if "type" not in values:
            if "properties" in values:
                values["type"] = "object"
            elif "items" in values:
                values["type"] = "array"

        # Get around pydantic bug wherein `Optional[conlists]`` throw a
        # 'NoneType' object is not iterable error if `None` is provided in init.
        for conlists in ("enum", "required"):
            if conlists in values and not values[conlists]:
                values.pop(conlists)

        return values

    @property
    def has_constraint(self) -> bool:
        """Return True if this schema has any constraints."""
        return bool(_CONSTRAINT_FIELDS & self.__fields_set__)

    @property
    def has_default(self) -> bool:
        """Return True if the schema has a default value."""
        return "default" in self.__fields_set__

    @property
    def python_type(self) -> Union[Type, List[Type]]:
        """Return Python type equivalent(s) for this schema (JSON) type."""
        if isinstance(self.type, list):
            return [_python_equivalent[t] for t in self.type]
        else:
            return _python_equivalent[self.type]

    @property
    def is_array(self) -> bool:
        """Return True if this schema is an array schema."""
        return self.items is not None or self.type == "array"

    @property
    def is_object(self) -> bool:
        """Return True if this schema is an object schema."""
        return (
            self.properties is not None
            or self.type == "object"
            and not self.all_of
            and not self.one_of
            and not self.any_of
            and not getattr(self, "ref", False)  # draft 6+
        )

    @property
    def json_validator(self) -> Type[Validator]:
        """Return jsonschema validator class for this schema.

        See also `validate_instance`.
        """
        if not hasattr(self, "_json_validator"):
            from jsonschema.validators import validator_for

            schema = self.dict(by_alias=True, exclude_unset=True)
            schema["$schema"] = self.schema_
            cls = validator_for(schema)
            cls.check_schema(schema)
            self._json_validator = cls(schema)
        return self._json_validator

    def validate_instance(self, instance: Any) -> dict:
        """Validate an object (instance) against this schema."""
        from jsonschema.exceptions import best_match

        error: ValidationError = best_match(self.json_validator.iter_errors(instance))
        if error is not None:
            raise error
        return instance


class Draft04JsonSchema(_JsonSchemaBase):
    """Model for Draft 4 JSON Schema."""

    schema_: str = Field("http://json-schema.org/draft-04/schema#", alias="$schema")
    id: Optional[str] = Field(None)
    exclusive_maximum: Optional[bool] = Field(None)
    exclusive_minimum: Optional[bool] = Field(None)
    required: Optional[StringArrayMin1] = Field(None)  # type: ignore
    dependencies: Optional[  # type: ignore
        Dict[str, Union[Draft04JsonSchema, StringArrayMin1]]
    ] = Field(None)

    # common to all schemas (could go in _JsonSchemaBase)
    # except we need the self-referrential type to be this class
    additional_items: Union[bool, Draft04JsonSchema, None] = Field(None)
    items: Union[Draft04JsonSchema, List[Draft04JsonSchema], None] = Field(None)
    additional_properties: Union[bool, Draft04JsonSchema, None] = Field(None)
    definitions: Optional[Dict[str, Draft04JsonSchema]] = Field(None)
    properties: Optional[Dict[str, Draft04JsonSchema]] = Field(None)
    pattern_properties: Optional[Dict[str, Draft04JsonSchema]] = Field(None)
    all_of: Optional[List[Draft04JsonSchema]] = Field(None)
    any_of: Optional[List[Draft04JsonSchema]] = Field(None)
    one_of: Optional[List[Draft04JsonSchema]] = Field(None)
    not_: Optional[Draft04JsonSchema] = Field(None, alias="not")


class _Draft06JsonSchema(_JsonSchemaBase):
    id: Optional[str] = Field(None, alias="$id")
    # ref: Optional[str] = Field(None, alias="$ref")
    examples: Optional[List[Any]] = Field(None)
    exclusive_maximum: Optional[float] = Field(None)
    exclusive_minimum: Optional[float] = Field(None)
    contains: Optional[Draft06JsonSchema] = Field(None)
    required: Optional[StringArray] = Field(None)  # type: ignore
    dependencies: Optional[  # type: ignore
        Dict[str, Union[Draft06JsonSchema, StringArray]]
    ] = Field(None)
    property_names: Optional[Draft06JsonSchema] = Field(None)
    const: Any = Field(None)


class Draft06JsonSchema(_Draft06JsonSchema):
    """Model for Draft 6 JSON Schema."""

    schema_: str = Field("http://json-schema.org/draft-06/schema#", alias="$schema")

    # common to all schemas (could go in _JsonSchemaBase)
    # except we need the self-referrential type to be this class
    # and... technically, all subschemas may also be booleans as of Draft 6,
    # not just additional_properties and additional_items
    additional_items: Union[bool, Draft06JsonSchema, None] = Field(None)
    items: Union[Draft06JsonSchema, List[Draft06JsonSchema], None] = Field(None)
    additional_properties: Union[bool, Draft06JsonSchema, None] = Field(None)
    # definitions: Optional[Dict[str, Draft06JsonSchema]] = Field(None)
    properties: Optional[Dict[str, Draft06JsonSchema]] = Field(None)
    pattern_properties: Optional[Dict[str, Draft06JsonSchema]] = Field(None)
    all_of: Optional[List[Draft06JsonSchema]] = Field(None)
    any_of: Optional[List[Draft06JsonSchema]] = Field(None)
    one_of: Optional[List[Draft06JsonSchema]] = Field(None)
    not_: Optional[Draft06JsonSchema] = Field(None, alias="not")


class Draft07JsonSchema(_Draft06JsonSchema):
    """Model for Draft 7 JSON Schema."""

    schema_: str = Field("http://json-schema.org/draft-07/schema#", alias="$schema")
    comment: Optional[str] = Field(None, alias="$comment")
    read_only: bool = Field(False)
    write_only: bool = Field(False)
    content_media_type: Optional[str] = Field(None)
    content_encoding: Optional[str] = Field(None)
    if_: Optional[Draft07JsonSchema] = Field(None, alias="if")
    then: Optional[Draft07JsonSchema] = Field(None)
    else_: Optional[Draft07JsonSchema] = Field(None, alias="else")

    # common to all schemas (could go in _JsonSchemaBase)
    # except we need the self-referrential type to be this class
    # and... technically, all subschemas may also be booleans as of Draft 6,
    # not just additional_properties and additional_items
    additional_items: Union[bool, Draft07JsonSchema, None] = Field(None)
    items: Union[Draft07JsonSchema, List[Draft07JsonSchema], None] = Field(None)
    additional_properties: Union[bool, Draft07JsonSchema, None] = Field(None)
    # definitions: Optional[Dict[str, Draft07JsonSchema]] = Field(None)
    properties: Optional[Dict[str, Draft07JsonSchema]] = Field(None)
    pattern_properties: Optional[Dict[str, Draft07JsonSchema]] = Field(None)
    all_of: Optional[List[Draft07JsonSchema]] = Field(None)
    any_of: Optional[List[Draft07JsonSchema]] = Field(None)
    one_of: Optional[List[Draft07JsonSchema]] = Field(None)
    not_: Optional[Draft07JsonSchema] = Field(None, alias="not")



# extra underscore in name to run this first

import pytest

from npe2 import (
    DynamicPlugin,
    PluginManager,
    io_utils,
    read,
    read_get_reader,
    write,
    write_get_writer,
)
from npe2.types import FullLayerData

SAMPLE_PLUGIN_NAME = "my-plugin"


def test_read(uses_sample_plugin):
    assert read(["some.fzzy"], stack=False) == [(None,)]


def test_read_with_unknown_plugin(uses_sample_plugin):
    # no such plugin name.... skips over the sample plugin & error is specific
    with pytest.raises(ValueError, match="Plugin 'nope' was selected"):
        read(["some.fzzy"], plugin_name="nope", stack=False)


def test_read_with_no_plugin():
    # no plugin passed and none registered
    with pytest.raises(ValueError, match="No readers returned"):
        read(["some.nope"], stack=False)


def test_read_return_reader(uses_sample_plugin):
    data, reader = read_get_reader("some.fzzy")
    assert data == [(None,)]
    assert reader.command == f"{SAMPLE_PLUGIN_NAME}.some_reader"


def test_read_return_reader_with_stack(uses_sample_plugin):
    data, reader = read_get_reader(["some.fzzy"], stack=True)
    assert data == [(None,)]
    assert reader.command == f"{SAMPLE_PLUGIN_NAME}.some_reader"


def test_read_list(uses_sample_plugin):
    data, reader = read_get_reader(["some.fzzy", "other.fzzy"])
    assert data == [(None,)]
    assert reader.command == f"{SAMPLE_PLUGIN_NAME}.some_reader"


null_image: FullLayerData = ([], {}, "image")


def test_writer_exec(uses_sample_plugin):
    # the sample writer knows how to handle two image layers
    result = write("test.tif", [null_image, null_image])
    assert result == ["test.tif"]

    result, contrib = write_get_writer("test.tif", [null_image, null_image])
    assert result == ["test.tif"]
    assert contrib.command == f"{SAMPLE_PLUGIN_NAME}.my_writer"


@pytest.mark.parametrize("layer_data", [[null_image, null_image], []])
def test_writer_exec_fails(layer_data, uses_sample_plugin):
    # the sample writer doesn't accept no extension
    with pytest.raises(ValueError):
        write("test/path", layer_data)


def test_writer_exec_fails2(uses_sample_plugin):
    # the sample writer doesn't accept 5 images
    with pytest.raises(ValueError):
        write("test.tif", [null_image, null_image, null_image, null_image, null_image])


def test_writer_single_layer_api_exec(uses_sample_plugin):
    # This writer doesn't do anything but type check.
    paths = write("test/path", [([], {}, "labels")])
    assert len(paths) == 1


def test_read_non_global_pm():
    pm = PluginManager()
    plugin = DynamicPlugin("my-plugin", plugin_manager=pm)

    @plugin.contribute.reader
    def read_path(path):
        def _reader(path):
            return [(None,)]

        return _reader

    assert io_utils._read(["some.fzzy"], stack=False, _pm=pm) == [(None,)]



from functools import partial
from importlib import metadata
from pathlib import Path
from unittest.mock import patch

import numpy as np
import pytest

from npe2 import PluginManager
from npe2.manifest import _npe1_adapter, utils
from npe2.manifest.contributions import SampleDataGenerator
from npe2.manifest.utils import SHIM_NAME_PREFIX


def test_adapter_no_npe1():
    pm = PluginManager()
    pm.discover()
    assert not pm._npe1_adapters


def test_npe1_adapter(uses_npe1_plugin, mock_cache: Path):
    """Test that the plugin manager detects npe1 plugins, and can index contribs"""
    pm = PluginManager()
    pm.discover(include_npe1=True)

    # we've found an adapter
    assert len(pm._npe1_adapters) == 1
    mf = pm.get_manifest("npe1-plugin")
    assert isinstance(mf, _npe1_adapter.NPE1Adapter)
    assert mf.package_metadata
    assert mf.package_metadata.version == "0.1.0"
    assert mf.package_metadata.name == "npe1-plugin"

    # it's currently unindexed and unstored
    assert not mf._cache_path().exists()
    assert not list(mock_cache.iterdir())

    with patch.object(
        _npe1_adapter,
        "manifest_from_npe1",
        wraps=_npe1_adapter.manifest_from_npe1,  # type: ignore
    ) as mock:
        pm.index_npe1_adapters()
        # the adapter has been cleared by the indexing
        assert len(pm._npe1_adapters) == 0
        # manifest_from_npe1 was called
        mock.assert_called_once_with(mf._dist, adapter=True)
        assert mf._cache_path().exists()
        # NOTE: accessing the `.contributions` object would have also triggered
        # importing, like pm.index_npe1_adapters() above, but it would not have
        # injected the contributions into the pm._contrib object.
        assert mf.contributions.sample_data

        mock.reset_mock()
        # clear and rediscover... this time we expect the cache to kick in
        pm.discover(clear=True, include_npe1=True)
        assert len(pm._npe1_adapters) == 1
        pm.index_npe1_adapters()
        assert len(pm._npe1_adapters) == 0
        mock.assert_not_called()


def test_npe1_adapter_cache(uses_npe1_plugin, mock_cache: Path):
    """Test that we can clear cache, etc.."""
    pm = PluginManager()
    pm.discover(include_npe1=True)

    with patch.object(
        _npe1_adapter,
        "manifest_from_npe1",
        wraps=_npe1_adapter.manifest_from_npe1,  # type: ignore
    ) as mock:
        # if we clear the cache, it should import again
        mf = pm.get_manifest("npe1-plugin")
        assert isinstance(mf, _npe1_adapter.NPE1Adapter)
        pm.index_npe1_adapters()
        mock.assert_called_once_with(mf._dist, adapter=True)
        assert mf._cache_path().exists()

        _npe1_adapter.clear_cache()
        assert not mf._cache_path().exists()

        mock.reset_mock()
        pm.discover(clear=True, include_npe1=True)
        pm.index_npe1_adapters()
        mf = pm.get_manifest("npe1-plugin")
        assert isinstance(mf, _npe1_adapter.NPE1Adapter)
        mock.assert_called_once_with(mf._dist, adapter=True)
        assert mf._cache_path().exists()
        _npe1_adapter.clear_cache(names=["not-our-plugin"])
        assert mf._cache_path().exists()
        _npe1_adapter.clear_cache(names=["npe1-plugin"])
        assert not mf._cache_path().exists()


def _get_mf() -> _npe1_adapter.NPE1Adapter:
    pm = PluginManager.instance()
    pm.discover(include_npe1=True)
    pm.index_npe1_adapters()
    mf = pm.get_manifest("npe1-plugin")
    assert isinstance(mf, _npe1_adapter.NPE1Adapter)
    return mf


def test_adapter_pyname_sample_data(uses_npe1_plugin, mock_cache):
    """Test that objects defined locally in npe1 hookspecs can be retrieved."""
    mf = _get_mf()
    samples = mf.contributions.sample_data
    assert samples
    sample_generator = next(s for s in samples if s.key == "local_data")
    assert isinstance(sample_generator, SampleDataGenerator)

    ONES = np.ones((4, 4))
    with patch.object(utils, "_import_npe1_shim", wraps=utils._import_npe1_shim) as m:
        func = sample_generator.get_callable()
        assert isinstance(func, partial)  # this is how it was defined in npe1-plugin
        pyname = f"{SHIM_NAME_PREFIX}npe1_module:napari_provide_sample_data_1"
        m.assert_called_once_with(pyname)
        assert np.array_equal(func(), ONES)

    # test nested sample data too
    sample_generator = next(s for s in samples if s.display_name == "Some local ones")
    func = sample_generator.get_callable()
    assert np.array_equal(func(), ONES)


def test_adapter_pyname_dock_widget(uses_npe1_plugin, mock_cache):
    """Test that objects defined locally in npe1 hookspecs can be retrieved."""
    mf = _get_mf()
    widgets = mf.contributions.widgets
    assert widgets
    wdg_contrib = next(w for w in widgets if w.display_name == "Local Widget")

    with patch.object(utils, "_import_npe1_shim", wraps=utils._import_npe1_shim) as m:
        caller = wdg_contrib.get_callable()
        assert isinstance(caller, partial)
        assert "<locals>.local_widget" in caller.keywords["function"].__qualname__
        pyname = (
            f"{SHIM_NAME_PREFIX}npe1_module:napari_experimental_provide_dock_widget_2"
        )
        m.assert_called_once_with(pyname)

        m.reset_mock()
        wdg_contrib2 = next(
            w for w in widgets if w.display_name == "local function" and w.autogenerate
        )
        caller2 = wdg_contrib2.get_callable()
        assert isinstance(caller2, partial)
        assert "<locals>.local_function" in caller2.keywords["function"].__qualname__
        pyname = f"{SHIM_NAME_PREFIX}npe1_module:napari_experimental_provide_function_1"
        m.assert_called_once_with(pyname)


def test_adapter_error_on_import():
    class FakeDist(metadata.Distribution):
        def read_text(self, filename):
            if filename == "METADATA":
                return "Name: fake-plugin\nVersion: 0.1.0\n"

        def locate_file(self, *_):
            ...

    adapter = _npe1_adapter.NPE1Adapter(FakeDist())

    def err():
        raise ImportError("No package found.")

    with pytest.warns(UserWarning) as record:
        with patch.object(_npe1_adapter, "manifest_from_npe1", wraps=err):
            _ = adapter.contributions
    assert "Error importing contributions for" in str(record[0])


def test_adapter_cache_fail(uses_npe1_plugin, mock_cache):
    pm = PluginManager()
    pm.discover(include_npe1=True)
    mf = pm.get_manifest("npe1-plugin")

    def err(obj):
        raise OSError("Can't cache")

    with patch.object(_npe1_adapter.NPE1Adapter, "_save_to_cache", err):
        # shouldn't reraise the error
        _ = mf.contributions



import pytest

from npe2.manifest.contributions import ConfigurationContribution, ConfigurationProperty
from npe2.manifest.contributions._json_schema import ValidationError

PROPS = [
    {
        "plugin.heatmap.location": {
            "type": "string",
            "default": "right",
            "enum": ["left", "right"],
            "enumDescriptions": [
                "Adds a heatmap indicator on the left edge",
                "Adds a heatmap indicator on the right edge",
            ],
        }
    }
]


@pytest.mark.parametrize("props", PROPS)
def test_config_contribution(props):
    cc = ConfigurationContribution(
        title="My Plugin",
        properties=props,
    )
    assert cc.title == "My Plugin"
    for key, val in cc.properties.items():
        assert val.dict(exclude_unset=True, by_alias=True) == props[key]


def test_warn_on_refs_defs():
    with pytest.warns(UserWarning):
        ConfigurationProperty(
            type="string",
            default="baz",
            description="quux",
            ref="http://example.com",
        )


CASES = [
    (
        {
            "type": str,
            "pattern": "^(\\([0-9]{3}\\))?[0-9]{3}-[0-9]{4}$",
            "pattern_error_message": "custom error",
        },
        "555-1212",
        "(888)555-1212 ext. 532",
    ),
    ({"type": "string", "minLength": 2}, "AB", "A"),
    ({"type": "string", "maxLength": 3}, "AB", "ABCD"),
    ({"type": "integer"}, 42, 3.123),
    ({"type": float}, 42.45, "3.123"),
    ({"type": int, "multipleOf": 10}, 30, 23),
    ({"type": "number", "minimum": 100}, 100, 99),
    ({"type": "number", "exclusiveMaximum": 100}, 99, 100),
    (
        {"properties": {"number": {"type": "number"}}},
        {"number": 1600},
        {"number": "1600"},
    ),
    (
        {
            "type": dict,
            "properties": {
                "number": {"type": "number"},
            },
            "additional_properties": False,
        },
        {"number": 1600},
        {"number": 1600, "street_name": "Pennsylvania"},
    ),
    ({"type": "array"}, [3, "diff", {"types": "of values"}], {"Not": "an array"}),
    ({"items": {"type": "number"}}, [1, 2, 3, 4, 5], [1, 2, "3", 4, 5]),
    (
        {
            "items": [
                {"type": "number"},
                {"type": "string"},
                {"enum": ["Street", "Avenue", "Boulevard"]},
                {"enum": ["NW", "NE", "SW", "SE"]},
            ]
        },
        [1600, "Pennsylvania", "Avenue", "NW"],
        [24, "Sussex", "Drive"],
    ),
    ({"type": [bool, int]}, True, "True"),
]


@pytest.mark.parametrize("schema, valid, invalid", CASES)
def test_config_validation(schema, valid, invalid):
    cfg = ConfigurationProperty(**schema)
    assert cfg.validate_instance(valid) == valid

    match = schema.get("pattern_error_message", None)
    with pytest.raises(ValidationError, match=match):
        assert cfg.validate_instance(invalid)

    assert cfg.is_array is ("items" in schema or cfg.type == "array")
    assert cfg.is_object is (cfg.type == "object")
    assert isinstance(cfg.has_constraint, bool)

    # check that we can can convert json type to python type
    for t in (
        cfg.python_type if isinstance(cfg.python_type, list) else [cfg.python_type]
    ):
        assert t.__module__ == "builtins"
    assert cfg.has_default is ("default" in schema)



from typing import TYPE_CHECKING, Any, Optional, Union

from pydantic import BaseModel, Extra, Field, validator

from npe2.manifest import _validators
from npe2.types import PythonName

from ._icon import Icon

if TYPE_CHECKING:
    from npe2._command_registry import CommandRegistry


class CommandContribution(BaseModel):
    """Contribute a **command** (a python callable) consisting of a unique `id`,
    a `title` and (optionally) a `python_path` that points to a fully qualified python
    callable.  If a `python_path` is not included in the manifest, it *must* be
    registered during activation with `register_command`.

    Note, some other contributions (e.g. `readers`, `writers` and `widgets`) will
    *point* to a specific command.  The command itself (i.e. the callable python
    object) will always appear in the `contributions.commands` section, but those
    contribution types may add additional contribution-specific metadata.

    ```{admonition} Future Plans
    Command contributions will eventually include an **icon**, **category**, and
    **enabled** state. Enablement is expressed with *when clauses*, that capture a
    conditional expression determining whether the command should be enabled or not,
    based on the current state of the program.  (i.e. "*If the active layer is a
    `Labels` layer*")

    Commands will eventually be availble in a Command Palette (accessible with a
    hotkey) but they can also show in other menus.
    ```
    """

    id: str = Field(
        ...,
        description="A unique identifier used to reference this command. While this may"
        " look like a python fully qualified name this does *not* refer to a python "
        "object; this identifier is specific to napari.  It must begin with "
        "the name of the package, and include only alphanumeric characters, plus "
        "dashes and underscores.",
    )
    _valid_id = validator("id", allow_reuse=True)(_validators.command_id)

    title: str = Field(
        ...,
        description="User facing title representing the command. This might be used, "
        "for example, when searching in a command palette. Examples: 'Generate lily "
        "sample', 'Read tiff image', 'Open gaussian blur widget'. ",
    )
    python_name: Optional[PythonName] = Field(
        None,
        description="Fully qualified name to a callable python object "
        "implementing this command. This usually takes the form of "
        "`{obj.__module__}:{obj.__qualname__}` "
        "(e.g. `my_package.a_module:some_function`)",
    )
    _valid_pyname = validator("python_name", allow_reuse=True)(_validators.python_name)

    short_title: Optional[str] = Field(
        None,
        description="Short title by which the command is represented in "
        "the UI. Menus pick either `title` or `short_title` depending on the context "
        "in which they show commands.",
    )
    category: Optional[str] = Field(
        None,
        description="Category string by which the command may be grouped in the UI.",
    )
    icon: Optional[Union[str, Icon]] = Field(
        None,
        description="Icon used to represent this command in the UI, on "
        "buttons or in menus. These may be [superqt](https://github.com/napari/superqt)"
        " fonticon keys, such as `'fa6s.arrow_down'`; though note that plugins are "
        "expected to depend on any fonticon libraries they use, e.g "
        "[fonticon-fontawesome6](https://github.com/tlambert03/fonticon-fontawesome6).",
    )
    enablement: Optional[str] = Field(
        None,
        description=(
            "Expression which must evaluate as true to enable the command in the UI "
            "(menu and keybindings). Does not prevent executing the command "
            "by other means, like the `execute_command` api."
        ),
    )

    class Config:
        extra = Extra.forbid

    def exec(
        self,
        args: tuple = (),
        kwargs: Optional[dict] = None,
        _registry: Optional["CommandRegistry"] = None,
    ) -> Any:
        if kwargs is None:
            kwargs = {}
        if _registry is None:
            from npe2._plugin_manager import PluginManager

            _registry = PluginManager.instance().commands
        return _registry.execute(self.id, args, kwargs)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from npe2 import implements
else:
    D = type("D", (), {"__getattr__": lambda *_: (lambda **_: (lambda f: f))})
    implements = D()


@implements.widget(id="some_widget", title="Create my widget", display_name="My Widget")
class SomeWidget:
    ...


@implements.sample_data_generator(
    id="my-plugin.generate_random_data",  # the plugin-name is optional
    title="Generate uniform random data",
    key="random_data",
    display_name="Some Random Data (512 x 512)",
)
def random_data():
    ...


@implements.widget(
    id="some_function_widget",
    title="Create widget from my function",
    display_name="A Widget From a Function",
    autogenerate=True,
)
def make_widget_from_function(x: int, threshold: int):
    ...



import sys
from unittest.mock import Mock, patch

import pytest

from npe2._command_registry import CommandHandler, CommandRegistry
from npe2._plugin_manager import PluginManager
from npe2.manifest.schema import PluginManifest
from npe2.types import PythonName

SAMPLE_PLUGIN_NAME = "my-plugin"


@pytest.fixture
def pm(sample_path):
    pm = PluginManager()
    pm.discover()
    assert len(pm._manifests) == 0
    sys.path.append(str(sample_path))
    try:
        pm.discover()
        yield pm
    finally:
        sys.path.remove(str(sample_path))


def test_discover_clear(uses_sample_plugin):
    pm = PluginManager.instance()
    assert SAMPLE_PLUGIN_NAME in pm._manifests
    reg_mock = Mock()
    pm.events.plugins_registered.connect(reg_mock)
    with patch.object(pm, "register", wraps=pm.register) as mock:
        pm.discover()
        mock.assert_not_called()  # nothing new to register
        reg_mock.assert_not_called()

        mock.reset_mock()
        pm.discover(clear=True)  # clear forces reregister
        mock.assert_called_once()
        reg_mock.assert_called_once_with({pm._manifests[SAMPLE_PLUGIN_NAME]})


def test_plugin_manager(pm: PluginManager):
    assert pm.get_command(f"{SAMPLE_PLUGIN_NAME}.hello_world")

    assert SAMPLE_PLUGIN_NAME not in pm._contexts
    ctx = pm.activate(SAMPLE_PLUGIN_NAME)
    assert SAMPLE_PLUGIN_NAME in pm._contexts
    assert pm.get_manifest(SAMPLE_PLUGIN_NAME)

    # dual activation is prevented
    assert pm.activate(SAMPLE_PLUGIN_NAME) is ctx

    assert pm.get_command(f"{SAMPLE_PLUGIN_NAME}.hello_world")

    assert pm.get_submenu("mysubmenu")
    with pytest.raises(KeyError):
        pm.get_submenu("not-a-submenu")
    assert len(list(pm.iter_menu("/napari/layer_context"))) == 2

    # deactivation
    assert SAMPLE_PLUGIN_NAME in pm._contexts
    pm.deactivate(SAMPLE_PLUGIN_NAME)
    assert SAMPLE_PLUGIN_NAME not in pm._contexts
    pm.deactivate(SAMPLE_PLUGIN_NAME)  # second time is a no-op
    assert SAMPLE_PLUGIN_NAME not in pm._contexts


def test_plugin_manager_register(sample_path):
    sys.path.append(str(sample_path))
    try:
        pm = PluginManager()
        pm.register(str(sample_path / "my_plugin" / "napari.yaml"))
        assert "my-plugin" in pm._manifests
        pm.unregister("my-plugin")
        assert "my-plugin" not in pm._manifests
        pm.register("my_plugin")
        assert "my-plugin" in pm._manifests
    finally:
        sys.path.remove(str(sample_path))


def test_plugin_manager_raises(pm: PluginManager):
    with pytest.raises(KeyError):
        pm.get_manifest("not-a-pluginxxx")
    with pytest.raises(KeyError):
        pm.activate("not a thing")
    with pytest.raises(KeyError):
        pm.get_command(f"{SAMPLE_PLUGIN_NAME}.not_a_thing")
    with pytest.raises(ValueError) as e:
        pm.register(PluginManifest(name=SAMPLE_PLUGIN_NAME))
    assert f"A manifest with name {SAMPLE_PLUGIN_NAME!r} already" in str(e.value)


def test_command_handler():
    with pytest.raises(RuntimeError):
        # cannot resolve something without either a python_name or function
        CommandHandler("hi").resolve()

    with pytest.raises(RuntimeError):
        # cannot resolve something without either a python_name or function
        CommandHandler("hi", python_name=PythonName("cannot.import.this")).resolve()


def test_command_reg_register():
    reg = CommandRegistry()
    with pytest.raises(ValueError):
        # must register non empty string id
        reg.register(1, lambda: None)  # type: ignore
    with pytest.raises(TypeError):
        # neither a string or a callable
        reg.register("other.id", 8)  # type: ignore

    with pytest.raises(ValueError):
        # must register non empty string id
        reg.register("some.id", "1_is_not.a_valid_python_name")
    reg.register("some.id", "this.is.a_valid_python_name")

    with pytest.raises(ValueError):
        # already registered
        reg.register("some.id", "this.is.a_valid_python_name")

    with pytest.raises(KeyError) as e:
        reg.get("not.a.command")
    assert "command 'not.a.command' not registered" in str(e.value)


def test_command_reg_get():
    def f(x, y):
        return x + y

    reg = CommandRegistry()
    reg.register("id", f)
    assert "id" in reg
    assert reg.get("id") is f
    assert reg.execute("id", (1, 2)) == 3


def _assert_sample_enabled(plugin_manager: PluginManager, enabled=True):
    i = SAMPLE_PLUGIN_NAME in plugin_manager._contrib._indexed
    assert i if enabled else not i

    _not = "not " if not enabled else ""
    # command
    if enabled:
        assert plugin_manager.get_command(f"{SAMPLE_PLUGIN_NAME}.hello_world")
    else:
        with pytest.raises(KeyError):
            assert plugin_manager.get_command(f"{SAMPLE_PLUGIN_NAME}.hello_world")

    # reader
    cmds = [r.command for r in plugin_manager.iter_compatible_readers("*.fzy")]
    b = f"{SAMPLE_PLUGIN_NAME}.some_reader" in cmds
    assert b if enabled else not b, f"Reader should {_not}be enabled"

    # writer
    cmds = [r.command for r in plugin_manager.iter_compatible_writers(["image"] * 2)]
    c = f"{SAMPLE_PLUGIN_NAME}.my_writer" in cmds
    assert c if enabled else not c, f"Writer should {_not}be enabled"

    d = "SampleTheme" in [t.label for t in plugin_manager.iter_themes()]
    assert d if enabled else not d, f"Theme should {_not}be enabled"


def test_enable_disable(uses_sample_plugin, plugin_manager: PluginManager, tmp_path):
    _assert_sample_enabled(plugin_manager)
    # just to test the enabled= kwarg on iter_manifests
    # (this would show *only* disabled plugins)
    assert not list(plugin_manager.iter_manifests(disabled=True))

    # Do disable
    mock = Mock()
    plugin_manager.events.enablement_changed.connect(mock)
    plugin_manager.disable(SAMPLE_PLUGIN_NAME)
    mock.assert_called_once_with({}, {SAMPLE_PLUGIN_NAME})  # enabled, disabled

    _assert_sample_enabled(plugin_manager, False)

    # stuff you can't do while disabled:
    with pytest.raises(ValueError):
        plugin_manager.activate(SAMPLE_PLUGIN_NAME)

    # re-enable
    mock.reset_mock()
    plugin_manager.enable(SAMPLE_PLUGIN_NAME)
    mock.assert_called_once_with({SAMPLE_PLUGIN_NAME}, {})  # enabled, disabled
    _assert_sample_enabled(plugin_manager)


def test_warn_on_register_disabled(uses_sample_plugin, plugin_manager: PluginManager):
    assert SAMPLE_PLUGIN_NAME in plugin_manager
    mf = plugin_manager[SAMPLE_PLUGIN_NAME]
    plugin_manager.disable(SAMPLE_PLUGIN_NAME)
    plugin_manager._manifests.pop(SAMPLE_PLUGIN_NAME)  # NOT good way to "unregister"
    with pytest.warns(UserWarning):
        plugin_manager.register(mf)


def test_plugin_manager_dict(uses_sample_plugin, plugin_manager: PluginManager):
    """Test exporting the plugin manager state with `dict()`."""
    d = plugin_manager.dict()
    assert SAMPLE_PLUGIN_NAME in d["plugins"]
    assert "disabled" in d
    assert "activated" in d

    d = plugin_manager.dict(
        include={"contributions", "package_metadata.version"},
        exclude={"contributions.writers", "contributions.readers"},
    )
    plugin_dict = d["plugins"][SAMPLE_PLUGIN_NAME]
    assert set(plugin_dict) == {"contributions", "package_metadata"}
    contribs = set(plugin_dict["contributions"])
    assert "readers" not in contribs
    assert "writers" not in contribs


def test_plugin_context_dispose():
    pm = PluginManager()
    mf = PluginManifest(name="test")
    pm.register(mf)
    mock = Mock()
    pm.get_context("test").register_disposable(mock)
    pm.deactivate("test")
    mock.assert_called_once()



from functools import wraps
from typing import List, Optional

from pydantic import Extra, Field

from npe2.manifest.utils import Executable, v2_to_v1
from npe2.types import ReaderFunction


class ReaderContribution(Executable[Optional[ReaderFunction]]):
    """Contribute a file reader.

    Readers may be associated with specific **filename_patterns** (e.g. "*.tif",
    "*.zip") and are invoked whenever `viewer.open('some/path')` is used on the
    command line, or when a user opens a file in the graphical user interface by
    dropping a file into the canvas, or using `File -> Open...`
    """

    command: str = Field(
        ..., description="Identifier of the command providing `napari_get_reader`."
    )
    filename_patterns: List[str] = Field(
        ...,
        description="List of filename patterns (for fnmatch) that this reader can "
        "accept. Reader will be tried only if `fnmatch(filename, pattern) == True`. "
        "Use `['*']` to match all filenames.",
    )
    accepts_directories: bool = Field(
        False, description="Whether this reader accepts directories"
    )

    class Config:
        extra = Extra.forbid

    def __hash__(self):
        return hash(
            (self.command, tuple(self.filename_patterns), self.accepts_directories)
        )

    def exec(self, *, kwargs):
        """
        We are trying to simplify internal npe2 logic to always deal with a
        (list[str], bool) pair instead of Union[PathLike, Seq[Pathlike]]. We
        thus wrap the Reader Contributions to still give them the old api. Later
        on we could add a "if manifest.version == 2" or similar to not have this
        backward-compatibility logic for new plugins.
        """
        kwargs = kwargs.copy()
        stack = kwargs.pop("stack", None)
        assert stack is not None
        kwargs["path"] = v2_to_v1(kwargs["path"], stack)
        callable_ = super().exec(kwargs=kwargs)

        if callable_ is None:  # pragma: no cover
            return None

        @wraps(callable_)
        def npe1_compat(paths, *, stack):
            path = v2_to_v1(paths, stack)
            return callable_(path)  # type: ignore

        return npe1_compat



import ast
from configparser import ConfigParser
from dataclasses import dataclass, field
from functools import cached_property
from importlib.metadata import EntryPoint
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

NPE1_EP = "napari.plugin"
NPE2_EP = "napari.manifest"


@dataclass
class PackageInfo:
    src_root: Optional[Path] = None
    package_name: str = ""
    entry_points: List[EntryPoint] = field(default_factory=list)
    setup_cfg: Optional[Path] = None
    setup_py: Optional[Path] = None
    pyproject_toml: Optional[Path] = None

    # @property
    # def packages(self) -> Optional[List[Path]]:
    #     return Path(self.top_module)

    @cached_property
    def _ep1(self) -> Optional[EntryPoint]:
        return next((ep for ep in self.entry_points if ep.group == NPE1_EP), None)

    @cached_property
    def _ep2(self) -> Optional[EntryPoint]:
        return next((ep for ep in self.entry_points if ep.group == NPE2_EP), None)

    @property
    def ep_name(self):
        if ep := self._ep1:
            return ep.name

    @property
    def ep_value(self):
        if ep := self._ep1:
            return ep.value

    @property
    def top_module(self) -> str:
        if ep := (self._ep1 or self._ep2):
            return ep.value.split(".", 1)[0].split(":", 1)[0]
        return ""  # pragma: no cover


def get_package_dir_info(path: Union[Path, str]) -> PackageInfo:
    """Attempt to *statically* get plugin info from a package directory."""
    path = Path(path).resolve()
    if not path.is_dir():  # pragma: no cover
        raise ValueError(f"Provided path is not a directory: {path}")

    info = PackageInfo(src_root=path)
    p = None

    # check for setup.cfg
    setup_cfg = path / "setup.cfg"
    if setup_cfg.exists():
        info.setup_cfg = setup_cfg
        p = ConfigParser()
        p.read(setup_cfg)
        info.package_name = p.get("metadata", "name", fallback="")
        if p.has_section("options.entry_points"):
            for group, val in p.items("options.entry_points"):
                name, _, value = val.partition("=")
                info.entry_points.append(EntryPoint(name.strip(), value.strip(), group))

    # check for setup.py
    setup_py = path / "setup.py"
    if setup_py.exists():
        info.setup_py = setup_py
        node = ast.parse(setup_py.read_text())
        visitor = _SetupVisitor()
        visitor.visit(node)
        if not info.package_name:
            info.package_name = visitor.get("name")
        if not info.entry_points:
            for group, vals in visitor.get("entry_points", {}).items():
                for val in vals if isinstance(vals, list) else [vals]:
                    name, _, value = val.partition("=")
                    info.entry_points.append(
                        EntryPoint(name.strip(), value.strip(), group)
                    )

    # # check for pyproject.toml
    # pyproject_toml = path / "pyproject.toml"
    # if pyproject_toml.exists():
    #     info.pyproject_toml = pyproject_toml
    #     with open(pyproject_toml, "r") as f:
    #         data = pytomlpp.load(f)
    #     if project := data.get("project"):
    #         info.package_name = project.get("name", "")
    #         for group, ep in project.get('entry-points', {}).items():
    #             for name, value in ep.items():
    #                 info.entry_points.append(EntryPoint(name, value, group))

    return info


class _SetupVisitor(ast.NodeVisitor):
    """Visitor to statically determine metadata from setup.py"""

    def __init__(self) -> None:
        super().__init__()
        self._names: Dict[str, Any] = {}
        self._setup_kwargs: Dict[str, Any] = {}

    def visit_Assign(self, node: ast.Assign) -> Any:
        if len(node.targets) == 1:
            target = node.targets[0]
            if isinstance(target, ast.Name) and isinstance(target.ctx, ast.Store):
                self._names[target.id] = self._get_val(node.value)

    def visit_Call(self, node: ast.Call) -> Any:
        if getattr(node.func, "id", "") == "setup":
            for k in node.keywords:
                key = k.arg
                value = self._get_val(k.value)
                self._setup_kwargs[str(key)] = value

    def _get_val(self, node: Optional[ast.expr]) -> Any:
        if isinstance(node, ast.Constant):
            return node.value
        if isinstance(node, ast.Name):
            return (
                self._names.get(node.id) if isinstance(node.ctx, ast.Load) else node.id
            )
        if isinstance(node, ast.Dict):
            keys = [self._get_val(k) for k in node.keys]
            values = [self._get_val(k) for k in node.values]
            return dict(zip(keys, values))
        if isinstance(node, ast.List):
            return [self._get_val(k) for k in node.elts]
        if isinstance(node, ast.Tuple):  # pragma: no cover
            return tuple(self._get_val(k) for k in node.elts)
        return str(node)  # pragma: no cover

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        return self._setup_kwargs.get(key, default)



from __future__ import annotations

from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from pydantic import BaseModel, ValidationError

from ._plugin_manager import PluginManager
from .manifest.contributions import (
    CommandContribution,
    ContributionPoints,
    ReaderContribution,
    SampleDataGenerator,
    WidgetContribution,
    WriterContribution,
)
from .manifest.schema import PluginManifest

C = TypeVar("C", bound=BaseModel)
T = TypeVar("T", bound=Callable[..., Any])


# a mapping of contribution type to string name in the ContributionPoints
# e.g. {ReaderContribution: 'readers'}
CONTRIB_NAMES = {v.type_: k for k, v in ContributionPoints.__fields__.items()}
for key in list(CONTRIB_NAMES):
    if getattr(key, "__origin__", "") == Union:
        v = CONTRIB_NAMES.pop(key)
        for t in key.__args__:
            CONTRIB_NAMES[t] = v


class DynamicPlugin:
    """A context manager that creates and modifies temporary plugin contributions.

    Parameters
    ----------
    name : str
        Optional name for this temporary plugin., by default "temp-plugin"
    plugin_manager : Optional[PluginManager]
        A plugin manager instance with which to associate this plugin. If `None` (the
        default), the global `PluginManager.instance()` will be used.
    manifest: Optional[PluginManifest],
        Optionally provide a manifest to use for this plugin. If not provided, a new
        manifest will be created.

    Examples
    --------
    >>> with TemporaryPlugin('name') as p:
    >>>     @p.contribute.sample_data
    >>>     def make_data() -> np.ndarray: ...
    """

    def __init__(
        self,
        name: str = "temp-plugin",
        plugin_manager: Optional[PluginManager] = None,
        manifest: Optional[PluginManifest] = None,
    ) -> None:
        if isinstance(manifest, PluginManifest):
            self.manifest = manifest
        else:
            self.manifest = PluginManifest(name=name)
        self.contribute = ContributionDecorators(self)
        self._pm = plugin_manager

    @property
    def name(self) -> str:
        """Name of the plugin."""
        return self.manifest.name

    @property
    def display_name(self) -> str:
        """Display name of the plugin."""
        return self.manifest.display_name

    def cleanup(self) -> None:
        """Remove this plugin from its plugin manager."""
        self.plugin_manager.unregister(self.manifest.name)

    def register(self) -> None:
        """Remove this plugin from its plugin manager."""
        self.plugin_manager.register(self.manifest)

    def clear(self) -> None:
        """Clear contributions."""
        self.plugin_manager.deactivate(self.manifest.name)
        self.plugin_manager._contrib.remove_contributions(self.manifest.name)
        self.manifest.contributions = ContributionPoints()

    @property
    def plugin_manager(self) -> PluginManager:
        """Return the plugin manager this plugin is registered in.

        If unset, will use the global plugin manager instance.
        """
        return self._pm if self._pm is not None else PluginManager.instance()

    @plugin_manager.setter
    def plugin_manager(self, pm: Optional[PluginManager]) -> None:
        """Set the plugin manager this plugin is registered in."""
        if pm is self._pm:  # pragma: no cover
            return

        my_cmds: Dict[str, Callable] = {
            k: v.function
            for k, v in self.plugin_manager.commands._commands.items()
            if k.startswith(self.manifest.name) and v.function
        }
        self.cleanup()
        self._pm = pm
        self.register()
        for k, v in my_cmds.items():
            self.plugin_manager.commands.register(k, v)

    def __enter__(self) -> DynamicPlugin:
        self.register()
        return self

    def __exit__(self, *_) -> None:
        self.cleanup()

    def spawn(
        self,
        name: Optional[str] = None,
        plugin_manager: Optional[PluginManager] = None,
        register: bool = False,
    ) -> DynamicPlugin:
        """Create a new DynamicPlugin instance with the same plugin manager.

        Parameters
        ----------
        name : Optional[str]
            If not provided, will increment current name, by default None
        plugin_manager : Optional[PluginManager]
            Plugin manager, by default the same as this plugin's plugin manager
        register : bool
            Whether to register the new plugin, by default False

        Returns
        -------
        DynamicPlugin
            A new DynamicPlugin instance.
        """
        pm = plugin_manager or self.plugin_manager
        assert isinstance(pm, PluginManager), "plugin_manager must be a PluginManager"

        if name:
            assert name not in pm._manifests, f"name {name!r} already in plugin manager"
            _name = name
        else:
            i = 1
            while (_name := f"{self.name}-{i}") in pm._manifests:
                i += 1

        new = type(self)(_name, plugin_manager=pm)
        if register:
            new.register()
        return new


class ContributionDecorators:
    """A set of decorators that facilitate adding contributions to a TemporaryPlugin.

    Examples
    --------
    >>> with TemporaryPlugin('name') as p:
    >>>     @p.contribute.sample_data
    >>>     def make_data() -> np.ndarray: ...
    >>>
    """

    def __init__(self, plugin: DynamicPlugin) -> None:
        self.plugin = plugin
        self.command = ContributionDecorator(plugin, CommandContribution)
        self.reader = ContributionDecorator(plugin, ReaderContribution)
        self.writer = ContributionDecorator(plugin, WriterContribution)
        self.widget = ContributionDecorator(plugin, WidgetContribution)
        self.sample_data = ContributionDecorator(plugin, SampleDataGenerator)


class ContributionDecorator(Generic[C]):
    """An actual instance of a contribution decorator.

    This holds the logic for actually adding a decorated function as a contribution
    of a specific `contrib_type` to a temporary plugin.
    """

    def __init__(self, plugin: DynamicPlugin, contrib_type: Type[C]) -> None:
        self.plugin = plugin
        self.contrib_type = contrib_type
        self._contrib_name = CONTRIB_NAMES[self.contrib_type]

    @overload
    def __call__(self, func: T, **kwargs) -> T:
        ...

    @overload
    def __call__(self, func: Literal[None] = None, **kwargs) -> Callable[[T], T]:
        ...

    def __call__(
        self, func: Optional[T] = None, **kwargs
    ) -> Union[T, Callable[[T], T]]:
        """Decorate function as providing this contrubtion type.

        This is the actual decorator used when one calls, eg.
        >>> @npe2plugin.contribute.reader
        >>> def some_func(path):
        >>>     ...
        """

        def _mark_contribution(_func: T, _kwargs=kwargs) -> T:
            try:
                self._set_defaults(_func, _kwargs)
                _kwargs = self._store_command(_func, _kwargs)
                self._store_contrib(_kwargs)
                self.plugin.plugin_manager._contrib.reindex(self._mf)
            except ValidationError as e:
                # cleanup any added commands
                if "command" in _kwargs:
                    new = [c for c in self.commands if c.id != _kwargs["command"]]
                    self._mf.contributions.commands = new
                    self.plugin.plugin_manager.commands.unregister(_kwargs["command"])
                raise AssertionError(
                    f"Invalid decorator for {self.contrib_type.__name__}.\n{e}"
                ) from e

            return _func

        return _mark_contribution if func is None else _mark_contribution(func)

    def _set_defaults(self, _func: T, kwargs: dict) -> None:
        """Populate contribution kwargs with reasonable type-specific defaults"""
        if self.contrib_type is ReaderContribution:
            kwargs.setdefault("filename_patterns", ["*"])
        if self.contrib_type is SampleDataGenerator:
            kwargs.setdefault("key", _func.__name__)
            kwargs.setdefault("display_name", _func.__name__)
        if self.contrib_type is WriterContribution:
            kwargs.setdefault("layer_types", [])

    def _store_contrib(self, kwargs: dict) -> None:
        """Store the new contribution in the manifest"""
        if self.contrib_type is not CommandContribution:
            self.contribution_list.append(self.contrib_type(**kwargs))

    def _store_command(self, func: T, kwargs: dict) -> dict:
        """Create a new command contribution for `func`"""
        kwargs.setdefault("title", func.__name__)
        kwargs.setdefault("id", f"{self.plugin.manifest.name}.{func.__name__}")
        cmd_kwargs = {
            k: kwargs.pop(k)
            for k in list(kwargs)
            if k in CommandContribution.__fields__
        }
        cmd = CommandContribution(**cmd_kwargs)
        self.commands.append(cmd)
        self.plugin.plugin_manager.commands.register(cmd.id, func)
        kwargs["command"] = cmd.id
        return kwargs

    @property
    def _mf(self) -> PluginManifest:
        """Return all contributions in currently in the temporary plugin"""
        return self.plugin.manifest

    @property
    def contribution_list(self) -> List[C]:
        """Return contributions of this type in the associated manifest."""
        if not getattr(self._mf.contributions, self._contrib_name):
            setattr(self._mf.contributions, self._contrib_name, [])
        return getattr(self._mf.contributions, self._contrib_name)

    @property
    def commands(self) -> List[CommandContribution]:
        """Return the CommandContributions in the associated manifest."""
        if not self._mf.contributions.commands:
            self._mf.contributions.commands = []
        return self._mf.contributions.commands



from importlib import metadata
from typing import Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Extra, Field, constr, root_validator
from pydantic.fields import SHAPE_LIST

# https://packaging.python.org/specifications/core-metadata/

MetadataVersion = Literal["1.0", "1.1", "1.2", "2.0", "2.1", "2.2"]
_alphanum = "[a-zA-Z0-9]"
PackageName = constr(regex=f"^{_alphanum}[a-zA-Z0-9._-]*{_alphanum}$")


class PackageMetadata(BaseModel):
    """Pydantic model for standard python package metadata.

    https://www.python.org/dev/peps/pep-0566/
    https://packaging.python.org/specifications/core-metadata/

    The `importlib.metadata` provides the `metadata()` function,
    but it returns a somewhat awkward `email.message.Message` object.
    """

    class Config:
        extra = Extra.ignore

    metadata_version: MetadataVersion = Field(
        "1.0", description="Version of the file format"
    )
    name: PackageName = Field(  # type: ignore
        ...,
        description="The name of the distribution. The name field "
        "is the primary identifier for a distribution.",
    )
    # technically there is PackageVersion regex at
    # https://www.python.org/dev/peps/pep-0440/#id81
    # but it will fail on some dev versions, and it's not worth it.
    version: str = Field(
        ...,
        description="A string containing the distribution's version number. "
        "This field must be in the format specified in PEP 440.",
    )
    dynamic: Optional[List[str]] = Field(
        None,
        description="A string containing the name of another core metadata "
        "field. The field names Name and Version may not be specified in this field.",
        min_ver="2.2",
    )
    platform: Optional[List[str]] = Field(
        None,
        description="A Platform specification describing an operating system "
        "supported by the distribution which is not listed in the “Operating System” "
        "Trove classifiers. See “Classifier” below.",
    )
    supported_platform: Optional[List[str]] = Field(
        None,
        description="Binary distributions containing a PKG-INFO file will use the "
        "Supported-Platform field in their metadata to specify the OS and CPU for "
        "which the binary distribution was compiled",
        min_ver="1.1",
    )
    summary: Optional[str] = Field(
        None, description="A one-line summary of what the distribution does."
    )
    description: Optional[str] = Field(
        None,
        description="A longer description of the distribution that can "
        "run to several paragraphs.",
    )
    description_content_type: Optional[str] = Field(
        None,
        description="A string stating the markup syntax (if any) used in the "
        "distribution's description, so that tools can intelligently render the "
        "description. The type/subtype part has only a few legal values: "
        "text/plain, text/x-rst, text/markdown",
        min_ver="2.1",
    )
    keywords: Optional[str] = Field(
        None,
        description="A list of additional keywords, separated by commas, to be used "
        "to assist searching for the distribution in a larger catalog.",
    )
    home_page: Optional[str] = Field(
        None,
        description="A string containing the URL for the distribution's home page.",
    )
    download_url: Optional[str] = Field(
        None,
        description="A string containing the URL from which THIS version of the "
        "distribution can be downloaded.",
        min_ver="1.1",
    )
    author: Optional[str] = Field(
        None,
        description="A string containing the author's name at a minimum; "
        "additional contact information may be provided.",
    )
    author_email: Optional[str] = Field(
        None,
        description="A string containing the author's e-mail address. It can contain "
        "a name and e-mail address in the legal forms for a RFC-822 From: header.",
    )
    maintainer: Optional[str] = Field(
        None,
        description="A string containing the maintainer's name at a minimum; "
        "additional contact information may be provided.",
        min_ver="1.2",
    )
    maintainer_email: Optional[str] = Field(
        None,
        description="A string containing the maintainer's e-mail address. It can "
        "contain a name and e-mail address in the legal forms for a "
        "RFC-822 From: header.",
        min_ver="1.2",
    )
    license: Optional[str] = Field(
        None,
        description="Text indicating the license covering the distribution where the "
        "license is not a selection from the “License” Trove classifiers. See "
        "“Classifier” below. This field may also be used to specify a particular "
        "version of a license which is named via the Classifier field, or to "
        "indicate a variation or exception to such a license.",
    )
    classifier: Optional[List[str]] = Field(
        None,
        description="Each entry is a string giving a single classification value for "
        "the distribution. Classifiers are described in PEP 301, and the Python "
        "Package Index publishes a dynamic list of currently defined classifiers.",
        min_ver="1.1",
    )
    requires_dist: Optional[List[str]] = Field(
        None,
        description="The field format specification was relaxed to accept the syntax "
        "used by popular publishing tools. Each entry contains a string naming some "
        "other distutils project required by this distribution.",
        min_ver="1.2",
    )
    requires_python: Optional[str] = Field(
        None,
        description="This field specifies the Python version(s) that the distribution "
        "is guaranteed to be compatible with. Installation tools may look at this "
        "when picking which version of a project to install. "
        "The value must be in the format specified in Version specifiers (PEP 440).",
        min_ver="1.2",
    )
    requires_external: Optional[List[str]] = Field(
        None,
        description="The field format specification was relaxed to accept the syntax "
        "used by popular publishing tools. Each entry contains a string describing "
        "some dependency in the system that the distribution is to be used. This "
        "field is intended to serve as a hint to downstream project maintainers, and "
        "has no semantics which are meaningful to the distutils distribution.",
        min_ver="1.2",
    )
    project_url: Optional[List[str]] = Field(
        None,
        description="A string containing a browsable URL for the project and a label "
        "for it, separated by a comma.",
        min_ver="1.2",
    )
    provides_extra: Optional[List[str]] = Field(
        None,
        description="A string containing the name of an optional feature. Must be a "
        "valid Python identifier. May be used to make a dependency conditional on "
        "whether the optional feature has been requested.",
        min_ver="2.1",
    )

    # rarely_used
    provides_dist: Optional[List[str]] = Field(None, min_ver="1.2")
    obsoletes_dist: Optional[List[str]] = Field(None, min_ver="1.2")

    @root_validator(pre=True)
    def _validate_root(cls, values):
        if "metadata_version" not in values:
            fields = cls.__fields__
            mins = {
                fields[n].field_info.extra.get("min_ver", "1.0")
                for n in values
                if n in fields
            }
            values["metadata_version"] = str(max(float(x) for x in mins))
        return values

    @classmethod
    def for_package(cls, name: str) -> "PackageMetadata":
        """Get PackageMetadata from a package name."""
        return cls.from_dist_metadata(metadata.metadata(name))

    # note, the metadata.PackageMetadata hint is only valid for py 3.10
    # before that, it was email.message.Message
    @classmethod
    def from_dist_metadata(cls, meta: "metadata.PackageMetadata") -> "PackageMetadata":
        """Generate PackageMetadata from importlib.metadata.PackageMetdata."""
        manys = [f.name for f in cls.__fields__.values() if f.shape == SHAPE_LIST]
        d: Dict[str, Union[str, List[str]]] = {}
        # looks like py3.10 changed the public protocol of metadata.PackageMetadata
        # and they don't want you to rely on the Mapping interface... however, the
        # __iter__ method doesn't iterate key value pairs, just keys, and I can't figure
        # out how to get multi-valued fields from that (e.g. Classifier)
        # might need to change this in the future
        for key, value in meta.items():  # type: ignore
            key = _norm(key)
            if key in manys:
                d.setdefault(key, []).append(value)  # type: ignore
            else:
                d[key] = value
        return cls.parse_obj(d)

    def __hash__(self) -> int:
        return id(self)


def _norm(string: str) -> str:
    return string.replace("-", "_").replace(" ", "_").lower()



from typing import Optional, Union

from pydantic import BaseModel
from pydantic.fields import Field

from ._icon import Icon


class SubmenuContribution(BaseModel):
    id: str = Field(description="Identifier of the menu to display as a submenu.")
    label: str = Field(
        description="The label of the menu item which leads to this submenu."
    )
    icon: Optional[Union[str, Icon]] = Field(
        None,
        description=(
            "(Optional) Icon which is used to represent the command in the UI."
            " Either a file path, an object with file paths for dark and light"
            "themes, or a theme icon references, like `$(zap)`"
        ),
    )



import sys
from pathlib import Path

import pytest

from npe2 import PluginManifest

DOCS_DIR = Path(__file__).parent.parent / "_docs"


def test_example_manifest():
    example = DOCS_DIR / "example_manifest.yaml"
    assert PluginManifest.from_file(example)


@pytest.mark.github_main_only
def test_render_docs(tmp_path, monkeypatch):
    sys.path.append(str(DOCS_DIR.parent))
    from _docs.render import main

    assert not list(tmp_path.glob("*.md"))
    main(tmp_path)
    assert list(tmp_path.glob("*.md"))
